""" Metafile Lister.

    Copyright (c) 2009, 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""

import hashlib
import json
import logging
import sys

from pathlib import Path

import bencode

from pyrosimple.scripts.base import ScriptBase


class MetafileLister(ScriptBase):
    """List contents of a bittorrent metafile."""

    # argument description for the usage information
    ARGS_HELP = "<metafile>..."

    def add_options(self):
        """Add program options."""
        self.add_bool_option(
            "--reveal",
            help="show full announce URL including keys, as well as full piece information",
        )
        self.add_bool_option(
            "--raw", help="print the metafile's raw content in JSON format"
        )
        self.add_bool_option(
            "-V",
            "--skip-validation",
            help="show broken metafiles with an invalid structure",
        )
        self.add_value_option(
            "-o",
            "--output",
            "KEY1,KEY2.SUBKEY,...",
            action="append",
            default=[],
            help="select fields to print, output is separated by TABs;"
            " note that __file__ is the path to the metafile,"
            " __hash__ is the info hash,"
            " and __size__ is the data size in bytes",
        )
        self.add_value_option(
            "-c",
            "--check-data",
            "PATH",
            help="check the hash against the data in the given path",
        )

    def mainloop(self):
        """The main loop."""
        from pyrosimple.util import metafile  # pylint: disable=import-outside-toplevel

        if not self.args:
            self.parser.print_help()
            self.parser.error("No metafiles given, nothing to do!")
            self.parser.exit()

        for idx, filename in enumerate(self.args):
            if idx and not self.options.output and not self.options.raw:
                print()
                print("~" * 79)

            try:
                # Read and check metafile
                try:
                    filename = Path(filename)
                    torrent = metafile.Metafile.from_file(filename)
                    if not self.options.skip_validation:
                        torrent.check_meta()
                except OSError as exc:
                    self.fatal(
                        "Can't read '%s' (%s)"
                        % (
                            filename,
                            str(exc).replace(": '%s'" % filename, ""),
                        )
                    )
                    raise

                if self.options.check_data:
                    # pylint: disable=import-outside-toplevel
                    from pyrosimple.util.ui import HashProgressBar

                    with HashProgressBar() as pb:
                        if (
                            logging.getLogger().isEnabledFor(logging.WARNING)
                            and sys.stdout.isatty()
                        ):
                            c = pb()
                            progress_callback = c.progress_callback
                        else:
                            progress_callback = None

                        torrent.hash_check(
                            Path(self.options.check_data),
                            progress_callback=progress_callback,
                        )
                listing = None

                if self.options.raw:
                    if not self.options.reveal and "info" in torrent:
                        # Shorten useless binary piece hashes
                        torrent["info"]["pieces"] = "<%d piece hashes>" % (
                            len(torrent["info"]["pieces"])
                            / len(hashlib.sha1().digest())
                        )

                    class BencodeJSONEncoder(json.JSONEncoder):
                        """Small helper class to translate 'binary' strings"""

                        def default(self, o):
                            if isinstance(o, bytes):
                                return o.hex().upper()
                            return super().default(o)

                    listing = BencodeJSONEncoder(indent=2).encode(torrent)
                elif self.options.output:

                    def splitter(fields):
                        "Yield single names for a list of comma-separated strings."
                        for flist in fields:
                            for field in flist.split(","):
                                yield field.strip()

                    data = {}
                    data["__file__"] = filename
                    if "info" in torrent:
                        data["__hash__"] = torrent.info_hash()
                        data["__size__"] = torrent.data_size()
                    values = []
                    for field in splitter(self.options.output):
                        try:
                            if field in data:
                                val = data[field]
                            else:
                                val = dict(torrent)
                                for key in field.split("."):
                                    val = val[key]
                        except KeyError:
                            self.LOG.error("%s: Field %r not found", filename, field)
                            break
                        else:
                            values.append(str(val))
                    else:
                        listing = "\t".join(values)
                else:
                    listing = "\n".join(torrent.listing(masked=not self.options.reveal))
            except (ValueError, KeyError, bencode.BencodeDecodeError) as exc:
                self.LOG.error(
                    "Bad metafile %r (%s: %s)", filename, type(exc).__name__, exc
                )
                raise
            else:
                if listing is not None:
                    print(listing)


def run():  # pragma: no cover
    """The entry point."""
    ScriptBase.setup()
    MetafileLister().run()


if __name__ == "__main__":
    run()



"""Move a torrent to a remote host

Currently uses consistent hashing to determine which remote host
gets the torrent (if multiple are specified)."""
import hashlib

from typing import Dict, List

from pyrosimple import config
from pyrosimple.job import base
from pyrosimple.util import rpc


def nodes_by_hash_weight(meta_id: str, nodes: List[str]) -> Dict[str, int]:
    """Weight nodes by hashing the meta_id"""
    result = {
        n: int.from_bytes(hashlib.sha256(meta_id.encode() + n.encode()).digest(), "big")
        for n in nodes
    }
    return dict(sorted(result.items(), key=lambda item: item[1]))


class Mover(base.MatchableJob):
    """Move torrent to remote host(s)"""

    def __init__(self, job_config=None):
        """Initalize torrent mover job"""
        super().__init__(job_config)
        self.config["hosts"] = [
            config.lookup_connection_alias(h) for h in self.config["hosts"]
        ]

    def run_item(self, item):
        """Statistics logger job callback."""
        for host in nodes_by_hash_weight(item.hash + item.alias, self.config["hosts"]):
            rproxy = rpc.RTorrentProxy(host)
            metahash = item.hash
            try:
                rproxy.d.hash(item.hash)
            except rpc.HashNotFound:
                pass
            else:
                self.log.info(
                    "Hash %s already exists at remote URL %s", item.hash, host
                )
                continue
            if self.config["dry_run"]:
                self.log.info("Would move %s to %s", metahash, rproxy.system.hostname())
                break
            item.move_to_host(host)
            self.log.info("Moved %s to %s", metahash, rproxy.system.hostname())
            break



""" RTorrent client proxy.

    Copyright (c) 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""


import base64
import functools
import json
import logging
import random
import urllib
import warnings

from typing import Any, Dict, List, Tuple, cast
from xmlrpc import client as xmlrpclib

from pyrosimple.io import scgi


logger = logging.getLogger(__name__)

NOHASH = (
    ""  # use named constant to make new-syntax commands with no hash easily searchable
)


class JSONRPCEncoder(json.JSONEncoder):
    """Encode xmlrpc.Binary data in a format jesec/rtorrent
    expects. The necessary command logic when handling load vs
    load.raw still needs to be handled separately."""

    def default(self, o):
        if isinstance(o, xmlrpclib.Binary):
            return "data:base64," + base64.b64encode(o.data).decode("ascii")
        return json.JSONEncoder.default(self, o)


CACHE_METHOD = {
    "d.chunk_size",
    "d.is_private",
    "d.name",
    "d.size_bytes",
    "d.size_chunks",
    "d.size_files",
    "system.api_version",
    "system.client_version",
    "system.library_version",
}


class RpcError(xmlrpclib.Fault):
    """Base class for XMLRPC protocol errors."""

    def __init__(self, faultString: str, faultCode: int = -500):
        super().__init__(faultCode, faultString)


class HashNotFound(RpcError):
    """Non-existing or disappeared hash."""

    def __init__(self, faultString: str, faultCode: int = -404):
        super().__init__(faultString, faultCode)


ERRORS = (RpcError,) + scgi.ERRORS


class RTorrentProxy(xmlrpclib.ServerProxy):
    # pylint: disable=super-init-not-called
    """Proxy to rTorrent's RPC interface.

    Method calls are built from attribute accesses, i.e. you can do
    something like `proxy.system.client_version()`.

    All methods from ServerProxy are being overridden due to the combination
    of self.__var name mangling and the __call__/__getattr__ magic.
    """

    def __init__(
        self,
        url,
        transport=None,
        encoding=None,
        verbose=False,
        allow_none=False,
        use_datetime=False,
        use_builtin_types=False,
        *,
        headers=(),
        context=None,
    ):
        # Get the url
        parsed_url = urllib.parse.urlsplit(url)
        queries = urllib.parse.parse_qs(parsed_url.query)
        # Config the connection details
        self.__rpc_codec = queries.get("rpc", ["xml"])[0]
        self.__url = url
        self.__host = parsed_url.netloc
        self.__handler = urllib.parse.urlunsplit(["", "", *parsed_url[2:]])
        # The /RPC2 convention just comes from the xmlrpc CLI tool, but is
        # generally used during ruTorrent setups
        if not self.__handler and parsed_url.scheme in ("http", "https"):
            warnings.warn(
                "Automatically adding '/RPC2' to the end of URLs is deprecated and will be removed in a future release."
                f"To fix this warning, change the connection URL to '{url}/RPC2'",
                stacklevel=2,
            )
            self.__handler = "/RPC2"
            self.__url += "/RPC2"

        if transport is None:
            if self.__rpc_codec == "json":
                codec = json
                headers = [("CONTENT_TYPE", "application/json")]
            elif self.__rpc_codec == "xml":
                codec = xmlrpclib
            handler = scgi.transport_from_url(url)
            transport = handler(
                url=self.__url,
                use_datetime=use_datetime,
                use_builtin_types=use_builtin_types,
                codec=codec,
                headers=headers,
            )
        self.__transport = transport
        self.__encoding = encoding or "utf-8"
        self.__verbose = verbose
        self.__allow_none = allow_none

    def __close(self):
        self.__transport.close()

    def __request_xml(self, methodname: str, params: Tuple[Any]):
        # Verbatim from parent method
        request = xmlrpclib.dumps(
            params,
            methodname,
            encoding=self.__encoding,
            allow_none=self.__allow_none,
        ).encode(self.__encoding, "xmlcharrefreplace")
        if self.__verbose:
            print("req: ", request)

        response = self.__transport.request(
            self.__host, self.__handler, request, verbose=self.__verbose
        )

        if len(response) == 1:
            return response[0]
        return response

    def __batch_request_json(self, calls: List) -> List:
        """Handle multicalls in place of XMLRPC's built-in
        system.multilcall."""
        batch_call: List[Dict] = [
            {
                "jsonrpc": "2.0",
                "method": call["methodName"],
                "params": call["params"] or [""],
                "id": index,
            }
            for index, call in enumerate(calls)
        ]
        request: bytes = json.dumps(batch_call).encode(
            self.__encoding, "xmlcharrefreplace"
        )
        response: Tuple[Dict] = cast(
            Tuple[Dict],
            self.__transport.request(
                self.__host,
                self.__handler,
                request,
                verbose=self.__verbose,
            ),
        )

        def sort_key(i: Dict) -> int:
            return int(i["id"])

        result = [[r["result"]] for r in sorted(response, key=sort_key)]
        return result

    def __request_json(self, methodname, params):
        if not params:
            params = [""]

        # system.multicall isn't defined by the app,
        # but by the xmlrpc library, so we intercept it
        # and turn it into a batch request
        if methodname == "system.multicall":
            return self.__batch_request_json(params[0])

        # This random ID feels silly but there's not much need for anything better at the moment
        # since the RPC interface is synchronous.
        rpc_id = random.randint(0, 100)
        request = (
            JSONRPCEncoder(separators=(",", ":"))
            .encode(
                {
                    "params": params,
                    "method": methodname,
                    "jsonrpc": "2.0",
                    "id": rpc_id,
                }
            )
            .encode(self.__encoding, "xmlcharrefreplace")
        )
        if self.__verbose:
            print("req: ", request)

        response: Dict = self.__transport.request(
            self.__host,
            self.__handler,
            request,
            verbose=self.__verbose,
        )

        if response["id"] != rpc_id:
            raise ValueError(
                f"RPC IDs do not match: sent={rpc_id} received={response['id']}"
            )
        if "error" in response:
            if "message" in response["error"]:
                if (
                    response["error"]["message"]
                    == "invalid parameters: info-hash not found"
                ):
                    raise HashNotFound(str(response["error"]))
                if "code" in response["error"]:
                    raise RpcError(
                        response["error"]["message"], response["error"]["code"]
                    )
            raise RpcError(f"Received error: {response['error']}")
        if "result" not in response:
            raise ValueError(f"Result not found in response: {response}")
        return response["result"]

    def __request(self, methodname, params):
        """Determines whether or not a request should be cached,
        then passes it to the appropriate method"""
        if methodname in CACHE_METHOD:
            return self.__cached_request(methodname, params)
        return self.__request_switch(methodname, params)

    @functools.lru_cache(maxsize=32)
    def __cached_request(self, methodname, params):
        """Simpled cache pass-through method to more easily take advantage of functools.lru_cache"""
        return self.__request_switch(methodname, params)

    def __request_switch(self, methodname, params):
        """Determines whether the request should go through
        xml or json and calls the appropriate method."""
        logger.debug("method '%s', params %s", methodname, params)
        try:
            if self.__rpc_codec == "xml":
                return self.__request_xml(methodname, params)
            if self.__rpc_codec == "json":
                return self.__request_json(methodname, params)
        except xmlrpclib.Fault as exc:
            if exc.faultString == "Could not find info-hash.":
                raise HashNotFound(  # pylint: disable=raise-missing-from
                    exc.faultString
                )
            raise exc
        raise ValueError(f"Invalid RPC protocol '{self.__rpc_codec}'")

    def __repr__(self):
        return f"<{self.__class__.__name__} via {self.__rpc_codec} for {self.__url}>"

    def __getattr__(self, name):
        # magic method dispatcher
        # Hardcode the most useful alias
        if name == "log":
            name = "print"
        return xmlrpclib._Method(self.__request, name)

    # note: to call a remote object with a non-standard name, use
    # result getattr(server, "strange-python-name")(args)

    def __call__(self, attr):
        """A workaround to get special attributes on the ServerProxy
        without interfering with the magic __getattr__
        """
        if attr == "close":
            return self.__close
        if attr == "transport":
            return self.__transport
        raise AttributeError(f"Attribute {attr} not found")

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.__close()



""" Metafile Editor.

    Copyright (c) 2010 The PyroScope Project <pyroscope.project@gmail.com>
"""


import copy
import hashlib
import os
import re
import time
import urllib.parse

from pathlib import Path

import bencode  # typing: ignore

from pyrosimple import config, error
from pyrosimple.scripts.base import ScriptBase
from pyrosimple.util import metafile


def replace_fields(meta, patterns):
    """Replace patterns in fields."""
    for pattern in patterns:
        try:
            field, regex, subst, _ = pattern.split(pattern[-1])

            # TODO: Allow numerical indices, and "+" for append
            namespace = meta
            keypath = [
                i.replace("\0", ".") for i in field.replace("..", "\0").split(".")
            ]
            for key in keypath[:-1]:
                namespace = namespace[key]

            namespace[keypath[-1]] = re.sub(regex, subst, namespace[keypath[-1]])
        except (KeyError, IndexError, TypeError, ValueError) as exc:
            raise error.UserError(f"Bad substitution '{pattern}' ({exc})!")

    return meta


class MetafileChanger(ScriptBase):
    """Change attributes of a bittorrent metafile."""

    # argument description for the usage information
    ARGS_HELP = "<metafile>..."

    # Keys of rTorrent session data
    RT_RESUME_KEYS = ("libtorrent_resume", "log_callback", "err_callback", "rtorrent")

    def add_options(self):
        """Add program options."""
        super().add_options()

        self.add_bool_option(
            "-n",
            "--dry-run",
            help="don't write changes to disk, just tell what would happen",
        )
        self.add_bool_option(
            "-V",
            "--no-skip",
            help="do not skip broken metafiles that fail the integrity check",
        )
        self.add_value_option(
            "-o",
            "--output-directory",
            "PATH",
            help="optional output directory for the modified metafile(s)",
        )
        self.add_bool_option(
            "-p", "--make-private", help="make torrent private (DHT/PEX disabled)"
        )
        self.add_bool_option(
            "-P", "--make-public", help="make torrent public (DHT/PEX enabled)"
        )
        self.add_value_option(
            "-s",
            "--set",
            "KEY=VAL",
            action="append",
            default=[],
            help="set a specific key to the given value; omit the '=' to delete a key",
        )
        self.add_value_option(
            "-r",
            "--regex",
            "KEY/REGEX/SUBST/",
            action="append",
            default=[],
            help="replace pattern in a specific key by the given substitution",
        )
        self.add_bool_option(
            "-C",
            "--clean",
            help="remove all non-standard data from metafile outside the info dict",
        )
        self.add_bool_option(
            "-A",
            "--clean-all",
            help="remove all non-standard data from metafile including inside the info dict",
        )
        self.add_bool_option(
            "-X",
            "--clean-xseed",
            help="like --clean-all, but keep libtorrent resume information",
        )
        self.add_bool_option(
            "-R",
            "--clean-rtorrent",
            help="remove all rTorrent session data from metafile",
        )
        self.add_value_option(
            "-H",
            "--hashed",
            "--fast-resume",
            "DATAPATH",
            help="add libtorrent fast-resume information (use {} in place of the torrent's name in DATAPATH)",
        )
        # TODO: chtor --tracker
        ##self.add_value_option("-T", "--tracker", "DOMAIN",
        ##    help="filter given torrents for a tracker domain")
        self.add_value_option(
            "-a",
            "--reannounce",
            "URL",
            help="set a new announce URL, but only if the old announce URL matches the new one",
        )
        self.add_value_option(
            "--reannounce-all",
            "URL",
            help="set a new announce URL on ALL given metafiles",
        )
        self.add_bool_option("--no-ssl", help="force announce URL to 'http'")
        self.add_bool_option(
            "--no-cross-seed",
            help="when using --reannounce-all, do not add a non-standard field to the info dict ensuring unique info hashes",
        )
        self.add_value_option(
            "--comment", "TEXT", help="set a new comment (an empty value deletes it)"
        )
        self.add_bool_option("--bump-date", help="set the creation date to right now")
        self.add_bool_option("--no-date", help="remove the 'creation date' field")

    def mainloop(self) -> None:
        """The main loop."""
        if not self.args:
            self.parser.print_help()
            self.parser.error("No metafiles given, nothing to do!")
            self.parser.exit()

        if 1 < sum(
            bool(i)
            for i in (
                self.options.no_ssl,
                self.options.reannounce,
                self.options.reannounce_all,
            )
        ):
            self.parser.error(
                "Conflicting options --no-ssl, --reannounce and --reannounce-all!"
            )

        # Set filter criteria for metafiles
        filter_url_prefix = None
        if self.options.reannounce:
            # <scheme>://<netloc>/<path>?<query>
            filter_url_prefix = urllib.parse.urlsplit(
                self.options.reannounce, allow_fragments=False
            )
            filter_url_prefix = urllib.parse.urlunsplit(
                (
                    filter_url_prefix.scheme,
                    filter_url_prefix.netloc,
                    "/",
                    "",
                    "",
                )
            )
            self.LOG.info(
                "Filtering for metafiles with announce URL prefix %r...",
                filter_url_prefix,
            )

        if self.options.reannounce_all:
            self.options.reannounce = self.options.reannounce_all
        else:
            # When changing the announce URL w/o changing the domain, don't change the info hash!
            self.options.no_cross_seed = True

        # Resolve tracker alias, if URL doesn't look like an URL
        if (
            self.options.reannounce
            and not urllib.parse.urlparse(self.options.reannounce).scheme
        ):
            for conn in config.multi_connection_lookup(self.options.reannounce):
                if urllib.parse.urlparse(conn).scheme:
                    self.options.reannounce = conn
                    break
            else:
                raise error.UserError(
                    "Unknown tracker alias or URL %r!" % (self.options.reannounce)
                )

        # go through given files
        bad = 0
        changed = 0
        for filename in self.args:
            try:
                # Read and remember current content
                torrent = metafile.Metafile.from_file(Path(filename))
            except (OSError, KeyError, bencode.BencodeDecodeError) as exc:
                self.LOG.warning(
                    "Skipping bad metafile %r (%s: %s)",
                    filename,
                    type(exc).__name__,
                    exc,
                )
                bad += 1
            else:
                # Check metafile integrity
                try:
                    torrent.check_meta()
                except ValueError as exc:
                    self.LOG.warning(
                        "Metafile %r failed integrity check: %s",
                        filename,
                        exc,
                    )
                    if not self.options.no_skip:
                        continue

                # Skip any metafiles that don't meet the pre-conditions
                if filter_url_prefix and not torrent["announce"].startswith(
                    filter_url_prefix
                ):
                    self.LOG.warning(
                        "Skipping metafile %r: not tracked by %r!",
                        filename,
                        filter_url_prefix,
                    )
                    continue

                # Keep resume info safe
                libtorrent_resume = {}
                if "libtorrent_resume" in torrent:
                    try:
                        libtorrent_resume["bitfield"] = torrent["libtorrent_resume"][
                            "bitfield"
                        ]
                    except KeyError:
                        pass  # nothing to remember

                    libtorrent_resume["files"] = copy.deepcopy(
                        torrent["libtorrent_resume"]["files"]
                    )

                # Change private flag?
                if self.options.make_private and not torrent["info"].get("private", 0):
                    self.LOG.info("Setting private flag...")
                    torrent["info"]["private"] = 1
                if self.options.make_public and torrent["info"].get("private", 0):
                    self.LOG.info("Clearing private flag...")
                    del torrent["info"]["private"]

                # Remove non-standard keys?
                if (
                    self.options.clean
                    or self.options.clean_all
                    or self.options.clean_xseed
                ):
                    torrent.clean_meta(
                        including_info=not self.options.clean,
                    )

                # Restore resume info?
                if self.options.clean_xseed:
                    if libtorrent_resume:
                        self.LOG.info("Restoring key 'libtorrent_resume'...")
                        torrent.setdefault("libtorrent_resume", {})
                        torrent["libtorrent_resume"].update(libtorrent_resume)
                    else:
                        self.LOG.warning("No resume information found!")

                # Clean rTorrent data?
                if self.options.clean_rtorrent:
                    for key in self.RT_RESUME_KEYS:
                        if key in torrent:
                            self.LOG.info("Removing key %r...", key)
                            del torrent[key]

                # Change announce URL?
                if self.options.reannounce:
                    torrent["announce"] = self.options.reannounce
                    if "announce-list" in torrent:
                        del torrent["announce-list"]

                    if not self.options.no_cross_seed:
                        # Enforce unique hash per tracker
                        torrent["info"]["x_cross_seed"] = hashlib.md5(
                            self.options.reannounce.encode()
                        ).hexdigest()
                if self.options.no_ssl:
                    # We're assuming here the same (default) port is used
                    torrent["announce"] = (
                        torrent["announce"]
                        .replace("https://", "http://")
                        .replace(":443/", ":80/")
                    )

                # Change comment or creation date?
                if self.options.comment is not None:
                    if self.options.comment:
                        torrent["comment"] = self.options.comment
                    elif "comment" in torrent:
                        del torrent["comment"]
                if self.options.bump_date:
                    torrent["creation date"] = int(time.time())
                if self.options.no_date and "creation date" in torrent:
                    del torrent["creation date"]

                # Add fast-resume data?
                if self.options.hashed:
                    datadir = self.options.hashed
                    if "{}" in datadir and not os.path.exists(datadir):
                        datadir = datadir.replace("{}", torrent["info"]["name"])
                    try:
                        torrent.add_fast_resume(datadir)
                    except OSError as exc:
                        self.fatal("Error making fast-resume data (%s)", exc)
                        raise

                # Set specific keys?
                torrent.assign_fields(self.options.set)
                replace_fields(torrent, self.options.regex)

                # Write new metafile, if changed

                if self.options.output_directory:
                    filename = os.path.join(
                        self.options.output_directory, os.path.basename(filename)
                    )
                    self.LOG.info("Will write %r...", filename)

                    if not self.options.dry_run:
                        with open(filename, "wb") as fh:
                            fh.write(bencode.bencode(torrent))
                        if "libtorrent_resume" in torrent:
                            # Also write clean version
                            filename = filename.replace(
                                ".torrent", "-no-resume.torrent"
                            )
                            del torrent["libtorrent_resume"]
                            self.LOG.info("Writing '%s'...", filename)
                            bencode.bwrite(torrent, filename)
                else:
                    self.LOG.info("Changing %r...", filename)

                    if not self.options.dry_run:
                        # Write to temporary file
                        tempname = os.path.join(
                            os.path.dirname(filename),
                            "." + os.path.basename(filename),
                        )
                        self.LOG.debug("Writing temporary file '%s'...", tempname)
                        torrent.save(Path(tempname))

                        try:
                            self.LOG.debug("Replacing file '%s'...", filename)
                            os.replace(tempname, filename)
                        except OSError as exc:
                            # TODO: Try to write directly, keeping a backup!
                            raise error.LoggableError(
                                "Can't rename tempfile %r to %r (%s)"
                                % (tempname, filename, exc)
                            )

                changed += 1

        # Print summary
        if changed:
            self.LOG.info(
                "%s %d metafile(s).",
                "Would've changed" if self.options.dry_run else "Changed",
                changed,
            )
        if bad:
            self.LOG.warning("Skipped %d bad metafile(s)!", bad)


def run():  # pragma: no cover
    """The entry point."""
    ScriptBase.setup()
    MetafileChanger().run()


if __name__ == "__main__":
    run()



# pylint: disable=attribute-defined-outside-init
""" rTorrent queue manager & daemon.

    Copyright (c) 2012 The PyroScope Project <pyroscope.project@gmail.com>
"""

import logging
import os
import signal
import sys
import time

from pathlib import Path
from typing import Dict

from apscheduler.schedulers.background import BackgroundScheduler
from daemon import DaemonContext
from daemon.pidfile import TimeoutPIDLockFile
from dynaconf.utils.boxing import DynaBox

from pyrosimple import config, error
from pyrosimple.scripts.base import ScriptBase, ScriptBaseWithConfig
from pyrosimple.util import logutil, pymagic


class RtorrentQueueManager(ScriptBaseWithConfig):
    """
    rTorrent queue manager & daemon.
    """

    # argument description for the usage information
    ARGS_HELP = ""

    POLL_TIMEOUT = 1.0

    RUNTIME_DIR = os.getenv("XDG_RUNTIME_DIR") or "~/.pyrosimple/run/"

    def add_options(self):
        """Add program options."""
        super().add_options()
        self.jobs: Dict = {}

        # basic options
        self.add_bool_option(
            "-n",
            "--dry-run",
            help="advise jobs not to do any real work, just tell what would happen",
        )
        self.add_bool_option(
            "--no-fork",
            "--fg",
            help="Don't fork into background (stay in foreground and log to console)",
        )
        self.add_value_option(
            "--run-once", "JOB", help="run the specified job once in the foreground"
        )
        self.add_bool_option("--stop", help="stop running daemon")
        self.add_bool_option(
            "--restart",
            help="stop any existing daemon, then start the process in the backgrounds",
        )
        self.add_bool_option("-?", "--status", help="Check daemon status")
        self.add_value_option(
            "--pid-file",
            "PATH",
            help="file holding the process ID of the daemon, when running in background",
        )

    # pylint: disable=no-self-use
    def parse_schedule(self, schedule):
        """Parse a job schedule."""
        result = {}

        for param in schedule.split():
            param = param.strip()
            try:
                key, val = param.split("=", 1)
                if key == "jitter":
                    val = int(val)
            except (TypeError, ValueError) as exc:
                raise error.ConfigurationError(
                    f"Bad param '{param}' in job schedule '{schedule}'"
                ) from exc
            else:
                result[key] = val

        return result

    def validate_config(self):
        """Handle and check configuration."""

        for name, params in config.settings.TORQUE.items():
            # Skip non-dictionary keys
            if not isinstance(params, DynaBox):
                continue
            for key in ("handler", "schedule"):
                if key not in params:
                    raise error.ConfigurationError(
                        f"Job '{name}' is missing the required '{key}' parameter"
                    )
            self.jobs[name] = dict(params)
            if self.options.dry_run:
                self.jobs[name]["dry_run"] = True
            if params.get("active", True):
                self.jobs[name]["handler"] = pymagic.import_name(params.handler)
            self.jobs[name]["schedule"] = self.parse_schedule(params.get("schedule"))

    def add_jobs(self):
        """Add configured jobs."""
        for name, params in self.jobs.items():
            if params.get("active", True):
                self.sched.add_job(
                    params["handler"](params, name).run,
                    name=name,
                    id=name,
                    trigger="cron",
                    **params["schedule"],
                )

    def reload_jobs(self):
        """Reload the configured jobs gracefully."""
        try:
            config.settings.configure()
            if self.running_config != dict(config.settings.TORQUE):
                self.LOG.info("Config change detected, reloading jobs")
                self.validate_config()
                self.sched.remove_all_jobs()
                self.add_jobs()
                self.running_config = dict(config.settings.TORQUE)
        except (Exception) as exc:  # pylint: disable=broad-except
            self.LOG.error("Error while checking config: %s", exc)

    def run_forever(self):
        """Run configured jobs until termination request."""
        self.running_config = dict(config.settings.TORQUE)
        while True:
            try:
                time.sleep(self.POLL_TIMEOUT)
                if config.settings.TORQUE.get("autoreload", False):
                    self.reload_jobs()
            except KeyboardInterrupt as exc:
                self.LOG.info("Termination request received (%s)", exc)
                break
            except SystemExit as exc:
                self.return_code = exc.code or 0
                self.LOG.info("System exit (RC=%r)", self.return_code)
                break

    def mainloop(self):
        """The main loop."""
        try:
            self.validate_config()
        except (error.ConfigurationError) as exc:
            self.fatal(exc)

        # Defaults for process control paths
        if not self.options.pid_file:
            self.options.pid_file = TimeoutPIDLockFile(
                Path(self.RUNTIME_DIR, "pyrotorque.pid").expanduser()
            )

        # Process control
        if self.options.status or self.options.stop or self.options.restart:
            if self.options.pid_file.is_locked():
                running, pid = True, self.options.pid_file.read_pid()
            else:
                running, pid = False, 0

            if self.options.status:
                if running:
                    print(f"Pyrotorque is running (PID {pid}).")
                    sys.exit(0)
                else:
                    print("No pyrotorque process found.")
                    sys.exit(1)

            if self.options.stop or self.options.restart:
                if running:
                    os.kill(pid, signal.SIGTERM)
                    self.LOG.debug("Process #%d sent SIGTERM.", pid)

                    # Wait for termination (max. 10 secs)
                    for _ in range(100):
                        if not self.options.pid_file.is_locked():
                            running = False
                            break
                        time.sleep(0.1)

                    self.LOG.info("Process #%d stopped.", pid)
                elif pid:
                    self.LOG.info("Process #%d NOT running anymore.", pid)
                else:
                    self.LOG.info(
                        "No pid file '%s'", (self.options.pid_file or "<N/A>")
                    )
            else:
                self.LOG.info(
                    "Process #%d %s running.", pid, "UP and" if running else "NOT"
                )

            if self.options.stop:
                self.return_code = error.EX_OK if running else error.EX_UNAVAILABLE
                return

        # Check if we only need to run once
        if self.options.run_once:
            params = self.jobs[self.options.run_once]
            if self.options.dry_run:
                params["dry_run"] = True
            params["handler_copy"] = params.get("handler")(params)
            params["handler_copy"].run()
            sys.exit(0)

        dcontext = DaemonContext(
            detach_process=False,
            stdin=sys.stdin,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
        # Detach, if not disabled via option
        if (
            not self.options.no_fork
        ):  # or getattr(sys.stdin, "isatty", lambda: False)():
            dcontext.detach_process = True
            dcontext.stdin = None
            dcontext.stderr = logutil.get_logfile()
            dcontext.stdout = logutil.get_logfile()
            dcontext.pidfile = self.options.pid_file
            self.LOG.info(
                "Writing pid to %s and detaching process...", self.options.pid_file
            )
            self.LOG.info("Logging stderr/stdout to %s", logutil.get_logfile())

        # Change logging format
        logging.basicConfig(
            force=True, format="%(asctime)s %(levelname)5s apscheduler: %(message)s"
        )

        with dcontext:
            # Set up services
            self.sched = BackgroundScheduler()

            # Run services
            self.sched.start()
            try:
                self.add_jobs()
                self.run_forever()
            finally:
                self.sched.shutdown()


def run():  # pragma: no cover
    """The entry point."""
    ScriptBase.setup()
    RtorrentQueueManager().run()


if __name__ == "__main__":
    run()



""" Administration Tool.

    Copyright (c) 2010 The PyroScope Project <pyroscope.project@gmail.com>
"""

import re

from datetime import datetime
from pathlib import Path
from xmlrpc import client as xmlrpclib

import pyrosimple

from pyrosimple import config
from pyrosimple.scripts.base import ScriptBase, ScriptBaseWithConfig
from pyrosimple.util import matching


class AdminTool(ScriptBaseWithConfig):
    """Support for administrative tasks."""

    # TODO: config create, dump, set, get
    # TODO: backup session/config

    def add_options(self):
        super().add_options()
        self.parser.set_defaults(func=None)
        subparsers = self.parser.add_subparsers()
        config_parser = subparsers.add_parser("config", help="Validate configuration")
        config_parser.set_defaults(func=self.config)
        config_parser.add_argument(
            "--check", help="Check config for any issues", action="store_true"
        )
        config_parser.add_argument(
            "--dump-rc", help="Print out the full rTorrent config", action="store_true"
        )
        backfill_parser = subparsers.add_parser(
            "backfill", help="Backfill missing custom fields from available data"
        )
        backfill_parser.set_defaults(func=self.backfill)
        backfill_parser.add_argument(
            "--dry-run",
            help="Print changes instead of applying them",
            action="store_true",
        )

    def dump_rc(self):
        """Print a representative .rtorrent.rc as gleaned from a running instance.

        This is neat but somewhat brittle, and behaves differently between XMLRPC and JSON-RPC."""

        proxy = pyrosimple.connect().open()
        methods = proxy.system.listMethods()
        # XXX This is a heuristic and might break in newer rTorrent versions!
        builtins = set(methods[: methods.index("view.sort_new") + 1])
        methods = set(methods)
        plain_re = re.compile(r"^[a-zA-Z0-9_.]+$")
        RC_CONTINUATION_THRESHOLD = 50

        def is_method(name):
            "Helper"
            prefixes = (
                "d.",
                "f.",
                "p.",
                "t.",
                "choke_group.",
                "session.",
                "system.",
                "throttle.",
                "trackers.",
                "ui.",
                "view.",
            )

            if name.endswith("="):
                name = name[:-1]
            return plain_re.match(name) and (
                name in methods or any(name.startswith(x) for x in prefixes)
            )

        def rc_quoted(text, in_brace=False):
            "Helper"
            if isinstance(text, list):
                wrap_fmt = "{%s}"
                try:
                    method_name = text[0] + ""
                except (TypeError, IndexError):
                    pass
                else:
                    if is_method(method_name):
                        wrap_fmt = "(%s)" if in_brace else "((%s))"
                        if (
                            ".set" not in method_name
                            and len(text) == 2
                            and text[1] == 0
                        ):
                            text = text[:1]
                text = wrap_fmt % ", ".join(
                    [rc_quoted(x, in_brace=(wrap_fmt[0] == "{")) for x in text]
                )
                return text.replace("))))", ")) ))")
            elif isinstance(text, int):
                return "(value, {:d})".format(text)
            elif plain_re.match(text) or is_method(text):
                return text
            else:
                return '"{}"'.format(text.replace("\\", "\\\\").replace('"', '\\"'))

        group = None
        for name in sorted(methods):
            try:
                value = proxy.method.get("", name)
                const = bool(proxy.method.const("", name))
            except xmlrpclib.Fault as exc:
                if "Key not found" in exc.faultString:
                    continue
                raise
            else:
                group, old_group = name.split(".", 1)[0], group
                if group == "event":
                    group = name
                if group != old_group:
                    print("")

                definition = None
                objtype = type(value)
                if objtype is list:
                    value = [rc_quoted(x) for x in value]
                    wrap_fmt = "((%s))" if value and is_method(value[0]) else "{%s}"
                    definition = wrap_fmt % ", ".join(value)
                elif objtype is dict:
                    print("method.insert = {}, multi|rlookup|static".format(name))
                    for key, val in sorted(value.items()):
                        val = rc_quoted(val)
                        if len(val) > RC_CONTINUATION_THRESHOLD:
                            val = "\\\n    " + val
                        print('method.set_key = {}, "{}", {}'.format(name, key, val))
                elif objtype is str:
                    definition = rc_quoted(value)
                elif objtype is int:
                    definition = "{:d}".format(value)
                else:
                    self.LOG.error(
                        "Cannot handle {!r} definition of method {}".format(
                            objtype, name
                        )
                    )
                    continue

                if definition:
                    if name in builtins:
                        print("{}.set = {}".format(name, definition))
                    else:
                        rctype = {str: "string", int: "value"}.get(objtype, "simple")
                        if const:
                            rctype += "|const"
                            const = None
                        if len(definition) > RC_CONTINUATION_THRESHOLD:
                            definition = "\\\n    " + definition
                        definition = definition.replace(
                            " ;     ", " ;\\\n     "
                        ).replace(",    ", ",\\\n    ")
                        print(
                            "method.insert = {}, {}, {}".format(
                                name, rctype, definition
                            )
                        )
                if const:
                    print("method.const.enable = {}".format(name))

    def backfill(self):
        """Backfill missing any missing metadata from available sources.
        Safe to run multiple times.
        """
        # pylint: disable=broad-except
        engine = pyrosimple.connect()
        engine.open()
        for i in engine.view("main", matching.create_matcher("loaded=0 metafile=/.+/")):
            try:
                mtime = int(Path(i.metafile).stat().st_mtime)
                if self.args.dry_run:
                    dt = datetime.fromtimestamp(mtime)
                    print(
                        f"Would set {i.hash} tm_loaded to {dt} from metafile {i.metafile}"
                    )
                else:
                    i.rpc_call("d.custom.set", ["tm_loaded", str(mtime)])
                    i.flush()
            except Exception as e:
                print(f"Could not set tm_loaded for {i.hash}: {e}")
        for i in engine.view("main", matching.create_matcher("loaded=0 path=/.+/")):
            try:
                mtime = int(Path(i.path).stat().st_mtime)
                if self.args.dry_run:
                    dt = datetime.fromtimestamp(mtime)
                    print(f"Would set {i.hash} tm_loaded to {dt} from path {i.path}")
                else:
                    i.rpc_call("d.custom.set", ["tm_loaded", str(mtime)])
                    i.flush()
            except Exception as e:
                print(f"Could not set tm_loaded for {i.hash}: {e}")
        for i in engine.view(
            "main", matching.create_matcher("completed=0 is_complete=yes path=/.+/")
        ):
            try:
                mtime = int(Path(i.path).stat().st_mtime)
                if self.args.dry_run:
                    dt = datetime.fromtimestamp(mtime)
                    print(f"Would set {i.hash} tm_completed to {dt} from path {i.path}")
                else:
                    i.rpc_call("d.custom.set", ["tm_completed", str(mtime)])
                    i.flush()
            except Exception as e:
                print(f"Could not set tm_loaded for {i.hash}: {e}")

    def config(self):
        """Handle the config subcommand"""
        if self.args.dump_rc:
            self.dump_rc()
        if self.args.check:
            try:
                config.autoload_scgi_url()
            except Exception:
                self.LOG.error("Error loading SCGI URL:")
                raise
            else:
                self.LOG.debug("Loaded SCGI URL successfully")
            try:
                pyrosimple.connect().open()
            except ConnectionRefusedError:
                self.LOG.error(
                    "SCGI URL '%s' found, but rTorrent may not be running!",
                    config.autoload_scgi_url(),
                )
                raise
            else:
                self.LOG.info("Connected to rTorrent successfully")

    def mainloop(self):
        self.args = self.parser.parse_args()
        if self.args.func is None:
            self.parser.print_help()
            return
        self.args.func()


def run():  # pragma: no cover
    """The entry point."""
    ScriptBase.setup()
    AdminTool().run()


if __name__ == "__main__":
    run()



"""Jobs for metric reporting"""
import threading

from collections import Counter
from time import sleep

from prometheus_client import start_http_server
from prometheus_client.core import REGISTRY, GaugeMetricFamily

from pyrosimple import config as config_ini
from pyrosimple import error
from pyrosimple.job.base import BaseJob
from pyrosimple.util import fmt, rpc
from pyrosimple.util.parts import Bunch


class EngineStats(BaseJob):
    """Simple rTorrent connection statistics logger."""

    def run(self):
        """Statistics logger job callback."""
        try:
            self.engine.open()
            self.log.info(
                "Stats for %s - up %s, %s torrents",
                self.engine.engine_id,
                fmt.human_duration(
                    self.engine.rpc.system.time() - self.engine.startup, 0, 2, True
                ).strip(),
                self.engine.rpc.view.size(rpc.NOHASH, "default"),
            )
        except (error.LoggableError, *rpc.ERRORS) as exc:
            self.log.warning(str(exc))


class ClientServer(threading.Thread):
    """Simple thread class to run the prometheus HTTP server
    in the background"""

    def __init__(self, port):
        super().__init__()
        self.port = int(port)

    def run(self):
        start_http_server(self.port)


class RtorrentCollector:
    """Base class for all the different collectors"""

    def __init__(self, proxy, config):
        self.proxy = proxy
        self.config = config
        self.prefix = self.config.get("prefix", "rtorrent_")

    def collect(self):
        """Must be implemented by the child classes"""
        raise NotImplementedError


class RtorrentItemCollector(RtorrentCollector):
    """Collects per-item information
    This will most likely caush prometheus to burn except with smaller instances"""

    def __init__(self, proxy, config):
        super().__init__(proxy, config)

        available_methods = set(self.proxy.system.listMethods())
        self.item_stat_methods = set(
            self.config.get("item-stats", ["d.down.total", "d.up.total"])
        )
        self.item_labels = set(self.config.get("item-labels", ["d.hash", "d.name"]))
        # Strip out unavailable methods
        self.item_stat_methods &= available_methods
        self.item_labels &= available_methods

    def collect(self):
        calls = [
            "d." + m + "=" for m in list(self.item_stat_methods) + self.item_labels
        ]
        result = self.proxy.d.multicall2("", "main", *calls)
        item_stats = {}
        for stat in self.item_stat_methods:
            item_stats[stat] = GaugeMetricFamily(
                self.prefix + stat.replace(".", "_"), stat, labels=self.item_labels
            )
        for i in result:
            info = dict(list(zip(list(self.item_stat_methods) + self.item_labels, i)))
            for stat, gauge in item_stats.items():
                gauge.add_metric([info[l] for l in self.item_labels], info[stat])
        for stat, guage in item_stats.items():
            yield guage


class RtorrentTrackerCollector(RtorrentCollector):
    """Collects tracker-based summaries"""

    def collect(self):
        tracker_gauge = GaugeMetricFamily(
            self.prefix + "tracker_amount",
            "Number of torrents belonging to a specific tracker",
            labels=["alias"],
        )
        tracker_error_gauge = GaugeMetricFamily(
            self.prefix + "tracker_errors",
            "Number of torrents with tracker errors belonging to a specific tracker",
            labels=["alias"],
        )

        item_fields = ["d.tracker_domain=", "d.message="]
        result = self.proxy.d.multicall2("", "main", *item_fields)

        trackers = Counter([config_ini.map_announce2alias(d[0]) for d in result])
        tracker_errors = Counter(
            [config_ini.map_announce2alias(d[0]) for d in result if d[1]]
        )

        for k, v in trackers.items():
            tracker_gauge.add_metric([k], v)
        for (
            k
        ) in (
            trackers.keys()
        ):  # Use the "trackers" keys to make sure all active trackers get a value
            tracker_error_gauge.add_metric([k], tracker_errors[k])

        yield tracker_gauge
        yield tracker_error_gauge


class RtorrentSystemCollector(RtorrentCollector):
    """Collects system information. This is both the most useful and least
    impactful collector, as it only takes a single system.multicall to collect
    all the information"""

    def __init__(self, proxy, config):
        super().__init__(proxy, config)
        stat_methods = [
            "network.http.current_open",
            "network.http.max_open",
            "network.max_open_files",
            "network.max_open_sockets",
            "network.open_files",
            "network.open_sockets",
            "network.total_handshakes",
            "pieces.memory.block_count",
            "pieces.memory.current",
            "pieces.memory.max",
            "pieces.memory.sync_queue",
            "pieces.preload.min_rate",
            "pieces.preload.min_size",
            "pieces.preload.type",
            "pieces.stats.total_size",
            "pieces.stats_not_preloaded",
            "pieces.stats_preloaded",
            "pieces.sync.queue_size",
            "startup_time",
            "system.files.closed_counter",
            "system.files.failed_counter",
            "system.files.opened_counter",
            "throttle.global_down.max_rate",
            "throttle.global_down.rate",
            "throttle.global_down.total",
            "throttle.global_up.max_rate",
            "throttle.global_up.rate",
            "throttle.global_up.total",
            "throttle.max_downloads.global",
            "throttle.max_uploads.global",
            "throttle.unchoked_downloads",
            "throttle.unchoked_uploads",
        ]

        self.info_methods = ["system.client_version", "system.library_version"]

        # Strip out unavailable methods
        self.system_stats = set(stat_methods) & set(self.proxy.system.listMethods())

    def collect(self):
        system_info = GaugeMetricFamily(
            self.prefix + "info",
            "rTorrent platform information",
            labels=[m.replace(".", "_") for m in self.info_methods],
        )
        system_view_size = GaugeMetricFamily(
            self.prefix + "view_size", "Size of rtorrent views", labels=["view"]
        )
        views = self.proxy.view.list()

        # Get data via multicall
        calls = (
            [dict(methodName=method, params=[]) for method in sorted(self.system_stats)]
            + [dict(methodName=method, params=[]) for method in self.info_methods]
            + [dict(methodName="view.size", params=["", view]) for view in views]
        )

        result = [r[0] for r in self.proxy.system.multicall(calls)]

        # Get numeric metrics
        for m in sorted(self.system_stats):
            yield GaugeMetricFamily(
                self.prefix + m.replace(".", "_"), m, value=result[0]
            )
            del result[0]

        # Get text-like information
        system_info.add_metric(result[0 : len(result) - len(views)], 1)
        yield system_info
        result = result[-len(views) :]

        # Get view information
        for v in views:
            system_view_size.add_metric([v], result[0])
            del result[0]
        yield system_view_size


class RtorrentExporter(BaseJob):
    """Expose rTorrent and host statistics for scraping by a Prometheus instance."""

    def __init__(self, config=None):
        """Set up RtorrentExporter."""
        super().__init__(config)
        self.prefix = self.config.get("prefix", "rtorrent_")
        self.proxy = self.engine.open()
        jobs = {
            "item": RtorrentItemCollector,
            "tracker": RtorrentTrackerCollector,
            "system": RtorrentSystemCollector,
        }
        for j in self.config.get("jobs", "system").split(","):
            j = j.strip()
            if j not in jobs:
                self.log.error("Job %s not found, skipping", j)
            else:
                REGISTRY.register(jobs[j](self.proxy, self.config))

        # Start the server right off the bat
        self.prom_thread = ClientServer(self.config.get("port", "8000"))
        self.prom_thread.start()

    def run(self):
        # NOOP, stats are generated at scrape time
        pass


if __name__ == "__main__":
    from pyrosimple import connect

    engine = connect()

    job = RtorrentExporter(Bunch(jobs="system", port=8005))
    while True:
        job.run()
        sleep(5)



"""Generic collections set to ease work with dicts"""
from collections import defaultdict


class Bunch(dict):
    """Generic attribute container that also is a dict."""

    def __getattr__(self, name):
        try:
            return dict.__getattribute__(self, name)
        except AttributeError:
            try:
                return self[name]
            except KeyError:
                # pylint: disable=raise-missing-from
                raise AttributeError(
                    "Bunch has no attribute %r in %s"
                    % (name, ", ".join([repr(i) for i in self.keys()]))
                )

    def __setattr__(self, name, value):
        self[name] = value

    def __repr__(self):
        return "Bunch(%s)" % ", ".join(sorted("%s=%r" % attr for attr in self.items()))


class DefaultBunch(Bunch, defaultdict):
    """Generic attribute container that also is a dict."""



""" Metafile Support.

    Copyright (c) 2009, 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""


import copy
import errno
import hashlib
import math
import os
import re
import time
import urllib

from pathlib import Path, PurePath
from typing import (
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import bencode  # typing: ignore

from pyrosimple import error
from pyrosimple.util import fmt, pymagic
from pyrosimple.util.parts import Bunch


ALLOWED_ROOT_NAME = re.compile(
    r"^[^/\\.~][^/\\]*$"
)  # cannot be absolute or ~user, and cannot have path parts
ALLOWED_PATH_NAME = re.compile(r"^(?:~\d+)?[^/\\~][^/\\]*$")


PASSKEY_RE = re.compile(r"(?<=[/=])[-_0-9a-zA-Z]{5,64}={0,3}(?=[/&]|$)")


PASSKEY_OK = (
    "announce",
    "TrackerServlet",
)


METAFILE_STD_KEYS = [
    _i.split(".")
    for _i in (
        "announce",
        "announce-list",  # BEP-0012
        "comment",
        "created by",
        "creation date",
        "encoding",
        "info",
        "info.length",
        "info.name",
        "info.piece length",
        "info.pieces",
        "info.private",
        "info.files",
        "info.files.length",
        "info.files.path",
    )
]

# PieceLogger and PieceFailer are both utility classes for passing into
# Metafile.make_info()'s piece_callback.
# Example usage:
class PieceLogger:
    """Holds some state to display useful error messages
    if pieces fail to hash check"""

    def __init__(self, meta, logger):
        self.piece_index = 0
        self.meta = meta
        self.log = logger

    def check_piece(self, filename, piece):
        "Callback for new piece"
        if (
            piece
            != self.meta["info"]["pieces"][self.piece_index : self.piece_index + 20]
        ):
            self.log.warning(
                "Piece #%d: Hashes differ in file %r",
                self.piece_index // 20,
                filename,
            )
        self.piece_index += 20


class PieceFailer(PieceLogger):
    """Raises an OSError if any pieces don't match"""

    def check_piece(self, filename, piece):
        "Callback for new piece"
        if (
            piece
            != self.meta["info"]["pieces"][self.piece_index : self.piece_index + 20]
        ):
            raise OSError(
                f"Piece #{self.piece_index // 20}: Hashes differ in file '{filename}'"
            )
        self.piece_index += 20


def mask_keys(announce_url: str) -> str:
    """Mask any passkeys (hex sequences) in an announce URL."""
    return PASSKEY_RE.sub(
        lambda m: m.group() if m.group() in PASSKEY_OK else "*" * len(m.group()),
        announce_url,
    )


class Metafile(dict):
    """A torrent metafile, representing structure and operations for a .torrent file."""

    @staticmethod
    def from_file(filename: Path):
        """Load a metafile directly from a file."""
        with filename.open("rb") as handle:
            raw_data = handle.read()
        return Metafile(bencode.decode(raw_data))

    def dict_copy(self) -> Dict:
        """Provide a copy of the metafile as a pure dict"""
        return copy.deepcopy(dict(self))

    def save(self, filename: Path) -> None:
        """Save the metafile to an actual file."""
        with filename.open("wb") as handle:
            handle.write(bencode.encode(dict(self)))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = pymagic.get_class_logger(self)
        self.ignore = []

    def check_info(self) -> None:
        """Validate info dict.

        Raise ValueError if validation fails.
        """
        info = self.get("info")
        if not isinstance(info, dict):
            raise ValueError("bad metainfo - not a dict")
        pieces = info.get("pieces")
        if not isinstance(pieces, bytes) or len(pieces) % 20 != 0:
            raise ValueError("bad metainfo - bad pieces key")

        piece_size = info.get("piece length")
        if not isinstance(piece_size, int) or piece_size <= 0:
            raise ValueError("bad metainfo - illegal piece length")

        name = info.get("name")
        if not isinstance(name, str):
            raise ValueError(
                f"bad metainfo - bad name (type is {type(name).__name__!r})"
            )
        if not ALLOWED_ROOT_NAME.match(name):
            raise ValueError(f"name {name} disallowed for security reasons")

        if ("files" in info) == ("length" in info):
            raise ValueError("single/multiple file mix")

        if "length" in info:
            length = info.get("length")
            if not isinstance(length, int) or length < 0:
                raise ValueError("bad metainfo - bad length")
        else:
            files = info.get("files")
            if not isinstance(files, (list, tuple)):
                raise ValueError("bad metainfo - bad file list")

            for item in files:
                if not isinstance(item, dict):
                    raise ValueError("bad metainfo - bad file value")

                length = item.get("length")
                if not isinstance(length, int) or length < 0:
                    raise ValueError("bad metainfo - bad length")

                path = item.get("path")
                if not isinstance(path, (list, tuple)) or not path:
                    raise ValueError("bad metainfo - bad path")

                for part in path:
                    if not isinstance(part, str):
                        raise ValueError("bad metainfo - bad path dir")
                    if part == "..":
                        raise ValueError(
                            "relative path in %s disallowed for security reasons"
                            % "/".join(path)
                        )
                    if part and not ALLOWED_PATH_NAME.match(part):
                        raise ValueError(f"path {part} disallowed for security reasons")

            file_paths = [os.sep.join(item["path"]) for item in files]
            if len(set(file_paths)) != len(file_paths):
                raise ValueError("bad metainfo - duplicate path")

    def check_meta(self) -> None:
        """Validate meta dict.

        Raise ValueError if validation fails.
        """
        if not isinstance(self, dict):
            raise ValueError("bad metadata - not a dictionary")
        if not isinstance(self.get("announce"), str):
            raise ValueError("bad announce URL - not a string")
        if not isinstance(self["info"], dict):
            raise ValueError("bad info key - not a dictionary")
        self.check_info()

    def info_hash(self) -> str:
        """Return info hash as a string."""
        return hashlib.sha1(bencode.encode(self["info"])).hexdigest().upper()

    def walk(self, datapath: Path) -> Generator[Path, None, None]:
        """Generate paths in "self.datapath", ignoring files/dirs as necesarry"""
        if datapath.is_dir():
            # Walk the directory tree
            for dirpath, dirnames, filenames in os.walk(datapath):
                # Don't scan blacklisted directories
                for bad in dirnames[:]:
                    if any(pattern.match(bad) for pattern in self.ignore):
                        dirnames.remove(bad)

                # Yield all filenames that aren't blacklisted
                for filename in filenames:
                    if not any(pattern.match(filename) for pattern in self.ignore):
                        yield Path(dirpath, filename)
        else:
            yield Path(datapath)

    def _calc_size(self, datapath) -> int:
        """Get total size of "self.datapath"."""
        return sum(os.path.getsize(filename) for filename in self.walk(datapath))

    def _make_info(
        self,
        files: Sequence[os.PathLike],
        piece_size: int,
        progress_callback: Optional[Callable[[int, int], None]] = None,
        piece_callback: Optional[Callable] = None,
        datapath: Optional[Path] = None,
    ):
        """Create info dict from a list of files."""
        # These collect the file descriptions and piece hashes
        file_list = []
        pieces = []

        # Initialize progress state
        hashing_secs = time.time()
        totalsize: int = sum([Path(filename).stat().st_size for filename in files])
        totalhashed: int = 0

        # Start a new piece
        sha1sum = hashlib.sha1()
        done: int = 0
        filename = None
        if datapath is None:
            datapath = os.path.commonpath(files)

        # Hash all files
        for filename in files:
            # Assemble file info
            filepath = Path(filename)
            filesize = filepath.stat().st_size
            rel_filepath = filepath.relative_to(datapath)
            file_list.append(
                {
                    "length": filesize,
                    "path": PurePath(rel_filepath).parts,
                }
            )
            self.log.debug("Hashing '%s', size %d...", filepath, filesize)

            # Open file and hash it
            fileoffset = 0
            with filepath.open("rb") as handle:
                while fileoffset < filesize:
                    # Read rest of piece or file, whatever is smaller
                    chunk = handle.read(min(filesize - fileoffset, piece_size - done))
                    sha1sum.update(chunk)
                    done += len(chunk)
                    fileoffset += len(chunk)
                    totalhashed += len(chunk)

                    # Piece is done
                    if done == piece_size:
                        pieces.append(sha1sum.digest())
                        if piece_callback:
                            piece_callback(filename, pieces[-1])

                        # Start a new piece
                        sha1sum = hashlib.sha1()
                        done = 0

                    # Report progress
                    if progress_callback:
                        progress_callback(totalhashed, totalsize)

        # Add hash of partial last piece
        if done > 0:
            pieces.append(sha1sum.digest())
            if piece_callback:
                piece_callback(filepath, pieces[-1])

        # Build the meta dict
        metainfo = {
            "pieces": b"".join(pieces),
            "piece length": piece_size,
            "name": os.path.basename(datapath),
        }

        # Handle directory/FIFO vs. single file
        if os.path.isdir(datapath):
            metainfo["files"] = file_list
        else:
            metainfo["length"] = totalhashed

        hashing_secs = time.time() - hashing_secs
        self.log.debug(
            "Hashing of %s took %.1f secs (%s/s)",
            fmt.human_size(totalhashed).strip(),
            hashing_secs,
            fmt.human_size(totalhashed / hashing_secs).strip(),
        )

        # Return validated info dict
        return metainfo, totalhashed

    def sanitize(self) -> Tuple[Set, Set]:
        """Try to fix common problems, especially transcode non-standard string encodings."""
        bad_encodings, bad_fields = set(), set()

        def sane_encoding(field, text):
            "Transcoding helper."
            if isinstance(text, str):
                return text.encode("utf-8")
            for encoding in ("utf-8", self.get("encoding", None), "cp1252"):
                if encoding:
                    try:
                        u8_text = text.decode(encoding).encode("utf-8")
                        if encoding != "utf-8":
                            bad_encodings.add(encoding)
                            bad_fields.add(field)
                        return u8_text
                    except UnicodeError:
                        continue
            # Broken beyond anything reasonable
            bad_encodings.add("UNKNOWN/EXOTIC")
            bad_fields.add(field)
            return str(text, "utf-8", "replace").replace("\ufffd", "_").encode("utf-8")

        # Go through all string fields and check them
        for field in ("comment", "created by"):
            if field in self.keys():
                self[field] = sane_encoding(field, self[field])

        self["info"]["name"] = sane_encoding("info name", self["info"]["name"])

        for entry in self["info"].get("files", []):
            entry["path"] = [sane_encoding("file path", i) for i in entry["path"]]

        return bad_encodings, bad_fields

    def assign_fields(self, assignments: List[str]) -> None:
        """Takes a list of C{key=value} strings and assigns them to the
        given metafile. If you want to set nested keys (e.g. "info.source"),
        you have to use a dot as a separator. For exotic keys *containing*
        a dot, double that dot ("dotted..key").

        Numeric values starting with "+" or "-" are converted to integers.

        If just a key name is given (no '='), the field is removed.
        """
        for assignment in assignments:
            try:
                val: Optional[Union[str, int]]
                if "=" in assignment:
                    field, val = assignment.split("=", 1)
                else:
                    field, val = assignment, None

                if val is not None and val[0] in "+-" and val[1:].isdigit():
                    val = int(val, 10)

                namespace = self
                # TODO: Allow numerical indices, and "+" for append
                keypath = [
                    i.replace("\0", ".") for i in field.replace("..", "\0").split(".")
                ]
                for key in keypath[:-1]:
                    # Create missing dicts as we go...
                    namespace = namespace.setdefault(key, {})
            except (KeyError, IndexError, TypeError, ValueError) as exc:
                raise error.UserError(
                    f"Bad assignment {assignment!r} ({exc})!"
                ) from exc
            else:
                if val is None:
                    del namespace[keypath[-1]]
                else:
                    namespace[keypath[-1]] = val

    def add_fast_resume(self, datapath: Path) -> None:
        """Add fast resume data to a metafile dict."""
        # Get list of files
        files = self["info"].get("files", None)
        single = files is None
        if single:
            if datapath.is_dir():
                datapath = datapath.joinpath(self["info"]["name"])
            files = [
                Bunch(
                    path=[os.path.abspath(datapath)],
                    length=self["info"]["length"],
                )
            ]

        # Prepare resume data
        resume = self.setdefault("libtorrent_resume", {})
        resume["bitfield"] = len(self["info"]["pieces"]) // 20
        resume["files"] = []
        piece_length = self["info"]["piece length"]
        offset = 0

        for fileinfo in files:
            # Get the path into the filesystem
            filepath = Path(*fileinfo["path"])
            if not single:
                filepath = Path(datapath, filepath)

            # Check file size
            if os.path.getsize(filepath) != fileinfo["length"]:
                raise OSError(
                    errno.EINVAL,
                    "File size mismatch for %r [is %d, expected %d]"
                    % (
                        filepath,
                        os.path.getsize(filepath),
                        fileinfo["length"],
                    ),
                )

            # Add resume data for this file
            resume["files"].append(
                dict(
                    priority=1,
                    mtime=int(os.path.getmtime(filepath)),
                    completed=(offset + fileinfo["length"] + piece_length - 1)
                    // piece_length
                    - offset // piece_length,
                )
            )
            offset += fileinfo["length"]
        self["libtorrent_resume"] = resume

    def data_size(self) -> int:
        """Calculate the size of a torrent based on parsed metadata."""
        info = self["info"]

        if "length" in info:
            # Single file
            return int(info["length"])
        # Directory structure
        return sum(f["length"] for f in info["files"])

    def _make_meta(
        self,
        datapath: Path,
        tracker_url: str,
        root_name: str,
        private: bool,
        progress: Optional[Callable[[int, int], None]] = None,
        piece_size: int = 0,
        piece_size_min: int = 2**15,
        piece_size_max: int = 2**24,
    ):
        """Create torrent dictionary from a file path."""
        if piece_size <= 0:
            # Calculate a good size for the data
            piece_size_exp = int(math.log(self._calc_size(datapath)) / math.log(2)) - 9
            # Limit it to the min and max
            piece_size = min(piece_size_max, max(piece_size_min, 2**piece_size_exp))
        # Round to the nearest power of two for all use cases
        piece_size = 2 ** (int(math.ceil(math.log(piece_size) / math.log(2))))

        # Build info hash
        info, totalhashed = self._make_info(
            sorted(self.walk(datapath)),
            piece_size,
            progress_callback=progress,
            datapath=datapath,
        )

        # Set private flag
        if private:
            info["private"] = 1

        # Freely chosen root name (default is basename of the data path)
        if root_name:
            info["name"] = root_name

        # Torrent metadata
        self["info"] = info
        self["announce"] = tracker_url.strip()

        # Return validated meta dict
        self.check_meta()
        return self, totalhashed

    def clean_meta(self, including_info: bool = False) -> Set[str]:
        """Clean meta dict.

        @param logger: If given, a callable accepting a string message.
        @return: Set of keys removed from C{meta}.
        """
        modified: Set[str] = set()

        for key in list(self.keys()):
            if [key] not in METAFILE_STD_KEYS:
                del self[key]
                modified.add(key)

        if including_info:
            for key in list(self["info"].keys()):
                if ["info", key] not in METAFILE_STD_KEYS:
                    del self["info"][key]
                    modified.add("info." + key)

            for entry in list(self["info"].get("files", [])):
                for key in list(entry.keys()):
                    if ["info", "files", key] not in METAFILE_STD_KEYS:
                        del entry[key]
                        modified.add("info.files." + key)

                # Remove crap that certain PHP software puts in paths
                entry["path"] = [i for i in entry["path"] if i]

        return modified

    @staticmethod
    def from_path(
        datapath,
        tracker_url,
        comment=None,
        root_name=None,
        created_by=None,
        private: bool = False,
        no_date: bool = False,
        progress=None,
        ignore=None,
        piece_size: int = 0,
        piece_size_min: int = 2**15,
        piece_size_max: int = 2**24,
    ):
        """Create a metafile with the path given on object creation.
        Returns the last metafile dict that was written (as an object, not bencoded).
        """
        # Lookup announce URLs from config file
        torrent = Metafile()
        if ignore:
            torrent.ignore = ignore
        try:
            if urllib.parse.urlparse(tracker_url).scheme:
                tracker_alias = (
                    urllib.parse.urlparse(tracker_url).netloc.split(":")[0].split(".")
                )
                tracker_alias = tracker_alias[-2 if len(tracker_alias) > 1 else 0]
            else:
                from pyrosimple import config  # pylint: disable=import-outside-toplevel

                tracker_alias, tracker_url = config.lookup_announce_url(tracker_url)
                tracker_url = tracker_url[0]
        except (KeyError, IndexError) as exc:
            raise error.UserError(
                f"Bad tracker URL {tracker_url!r}, or unknown alias!"
            ) from exc

        meta, _ = torrent._make_meta(
            datapath,
            tracker_url,
            root_name,
            private,
            progress,
            piece_size,
            piece_size_min,
            piece_size_max,
        )

        # Add optional fields
        if comment:
            meta["comment"] = comment
        if created_by:
            meta["created by"] = created_by
        if not no_date:
            meta["creation date"] = int(time.time())
        return Metafile(meta)

    def hash_check(
        self, datapath: Path, progress_callback=None, piece_callback=None
    ) -> bool:
        """Check piece hashes of a metafile against the given datapath."""

        if "length" in self["info"]:
            files = [datapath]
        else:
            files = [Path(datapath, *i["path"]) for i in self["info"]["files"]]
        datameta, _ = self._make_info(
            files,
            int(self["info"]["piece length"]),
            progress_callback=progress_callback,
            piece_callback=piece_callback,
        )
        return bool(datameta["pieces"] == self["info"]["pieces"])

    def listing(self, masked=True):
        """List torrent info & contents in human-readable format. Returns a list of formatted lines."""
        # Assemble data
        bad_encodings = []
        bad_fields = []
        announce = self["announce"]
        info = self["info"]
        infohash = self.info_hash()

        total_size = self.data_size()
        piece_length = info["piece length"]
        piece_number, last_piece_length = divmod(total_size, piece_length)

        # Build result
        result: List[str] = [
            f"NAME {self['info']['name']}",
            "SIZE %s (%i * %s + %s)"
            % (
                fmt.human_size(total_size).strip(),
                piece_number,
                fmt.human_size(piece_length).strip(),
                fmt.human_size(last_piece_length).strip(),
            ),
            "META %s (pieces %s %.1f%%)"
            % (
                fmt.human_size(self.data_size()).strip(),
                fmt.human_size(len(info["pieces"])).strip(),
                100.0 * len(info["pieces"]) / self.data_size(),
            ),
            f"HASH {infohash.upper()}",
            f"URL  {(mask_keys if masked else str)(announce)}",
            "PRV  %s"
            % (
                "YES (DHT/PEX disabled)"
                if info.get("private")
                else "NO (DHT/PEX enabled)"
            ),
            "TIME %s"
            % (
                "N/A"
                if "creation date" not in self
                else time.strftime(
                    "%Y-%m-%d %H:%M:%S", time.localtime(self["creation date"])
                )
            ),
        ]

        for label, key in (("BY  ", "created by"), ("REM ", "comment")):
            if key in self:
                result.append(f"{label} {self.get(key, 'N/A')}")

        result.extend(
            [
                "",
                "FILE LISTING%s"
                % ("" if "length" in info else " [%d file(s)]" % len(info["files"]),),
            ]
        )
        if "length" in info:
            # Single file
            result.append(
                "%-69s%9s"
                % (
                    info["name"],
                    fmt.human_size(total_size),
                )
            )
        else:
            # Directory structure
            result.append(f"{info['name']}/")
            oldpaths = [None] * 99
            for entry in info["files"]:
                # Remove crap that certain PHP software puts in paths
                entry_path = [i for i in entry["path"] if i]

                for idx, item in enumerate(entry_path[:-1]):
                    if item != oldpaths[idx]:
                        result.append(f"{' ' * (4 * (idx + 1))}{item}/")
                        oldpaths[idx] = item
                result.append(
                    "%-69s%9s"
                    % (
                        " " * (4 * len(entry_path)) + entry_path[-1],
                        fmt.human_size(entry["length"]),
                    )
                )

        if bad_encodings:
            result.extend(
                [
                    "",
                    "WARNING: Bad encoding(s) {} in these fields: {}".format(
                        ", ".join(sorted(bad_encodings)), ", ".join(sorted(bad_fields))
                    ),
                    "Use the --raw option to inspect these encoding issues.",
                ]
            )

        return result



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pyrosimple',
 'pyrosimple.io',
 'pyrosimple.job',
 'pyrosimple.scripts',
 'pyrosimple.torrent',
 'pyrosimple.ui',
 'pyrosimple.util']

package_data = \
{'': ['*'], 'pyrosimple': ['data/*']}

install_requires = \
['Jinja2>=3.1.0,<4.0.0',
 'bencode.py>=4.0.0,<5.0.0',
 'dynaconf[toml]>=3.1.0,<4.0.0',
 'parsimonious>=0.10.0,<0.11.0',
 'prometheus-client>=0.14.1,<0.15.0',
 'prompt-toolkit>=3.0.30,<4.0.0',
 'python-daemon>=2.3.0,<3.0.0',
 'requests>=2.28.1,<3.0.0',
 'shtab>=1.5.5,<2.0.0']

extras_require = \
{':python_version < "3.9"': ['importlib-resources>=5.4.0,<6.0.0'],
 'torque': ['APScheduler>=3.9.0,<4.0.0', 'pyinotify>=0.9.6,<0.10.0']}

entry_points = \
{'console_scripts': ['chtor = pyrosimple.scripts.chtor:run',
                     'lstor = pyrosimple.scripts.lstor:run',
                     'mktor = pyrosimple.scripts.mktor:run',
                     'pyroadmin = pyrosimple.scripts.pyroadmin:run',
                     'pyrotorque = pyrosimple.scripts.pyrotorque:run',
                     'rtcontrol = pyrosimple.scripts.rtcontrol:run',
                     'rtxmlrpc = pyrosimple.scripts.rtxmlrpc:run']}

setup_kwargs = {
    'name': 'pyrosimple',
    'version': '2.3.3',
    'description': 'A stripped-down version of the pyrocore tools for working with rTorrent',
    'long_description': "# pyrosimple\n\n[![GitHub Workflow Status](https://img.shields.io/github/workflow/status/kannibalox/pyrosimple/Pylint)](https://github.com/kannibalox/pyrosimple/actions/workflows/pylint.yml)\n[![PyPI](https://img.shields.io/pypi/v/pyrosimple)](https://pypi.org/project/pyrosimple/)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/pyrosimple)\n\nA overhauled Python 3 fork of the [pyrocore\ntools](https://github.com/pyroscope/pyrocore), for working with the\n[rTorrent client](https://github.com/rakshasa/rtorrent).\n\n## Installation\n\n```bash\npip install pyrosimple\n# pip install 'pyrosimple[torque]' # Optional dependencies for using pyrotorque\n```\n\nSee the [documentation for usage](https://kannibalox.github.io/pyrosimple/).\nIf you've used rtcontrol/rtxmlrpc before, you should feel right at home.\n\n## What's the point of this?\n\nThe pyrocore tools are great, but being stuck on python 2, along with\nthe complicated install procedure made integrating both the tools and\nthe code into other processes very painful.\n\n## Differences from pyrocore\n\nThe following lists are not exhaustive, and don't cover many of the\ninternal improvements and refactoring.\n\n- Only supports python 3 and rTorrent 0.9.6+ (this includes\n  rtorrent-ps and rtorrent-ps-ch)\n- Simpler poetry-based build system, with a single package to install\n- Performance improvements (faster templating and only fetching\n  required fields)\n\n### New features\n\n- Multi-instance support for rtcontrol/rtxmlrpc\n- Replaced Tempita with Jinja2\n- Support for JSON-RPC (only implemented by\n  https://github.com/jesec/rtorrent)\n- Actions to move torrent between paths, or torrents between hosts\n\n## Legacy branch\n\nIf you just want to use the pyrocore tools on python 3 without all the\nnew features, you can use the `release-1.X` branch (1.3 is the latest\nrelease at time of writing).  These releases will only receive bug\nfixes or changes to maintain compatibility with the original pyrocore\ntools.\n",
    'author': 'kannibalox',
    'author_email': 'kannibalox@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kannibalox/pyrosimple',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



""" Torrent Engine Interface.

    Copyright (c) 2009, 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""

import logging
import os
import re
import time
import warnings

from typing import Callable, Dict, Optional, Set, cast

from pyrosimple import config, error
from pyrosimple.util import fmt, matching, metafile, rpc, traits


logger = logging.getLogger(__name__)


def untyped(val):
    """A type specifier for fields that does nothing."""
    return val


def ratio_float(intval: float) -> float:
    """Convert scaled integer ratio to a normalized float."""
    return intval / 1000.0


def percent(floatval: float) -> float:
    """Convert float ratio to a percent value."""
    return floatval * 100.0


def _duration(start: Optional[float], end: Optional[float]) -> Optional[float]:
    """Return time delta."""
    if start is None:
        start = 0.0
    if end is None:
        end = 0.0
    if start and end:
        if start > end:
            return None
        return end - start
    if start:
        return time.time() - start
    return None


def _interval_split(
    interval: str, only: Optional[str] = None, event_re=re.compile("[A-Z][0-9]+")
):
    """Split C{interval} into a series of event type and timestamp tuples.
    An exaple of the input is "R1283008245P1283008268".
    Returns events in reversed order (latest first).
    """

    def split_event(event):
        "Helper to parse events."
        kind, val = event[:1], event[1:]
        try:
            return kind, float(val)
        except (TypeError, ValueError):
            return None, 0

    return list(
        reversed(
            [
                split_event(i)
                for i in event_re.findall(interval)
                if not only or i.startswith(only)
            ]
        )
    )


def _interval_sum(
    interval: str, start: Optional[float] = None, end: Optional[float] = None
) -> Optional[int]:
    """Return sum of intervals between "R"esume and "P"aused events
    in C{interval}, optionally limited by a time window defined
    by C{start} and C{end}. Return ``None`` if there's no sensible
    information.

    C{interval} is a series of event types and timestamps,
    e.g. "R1283008245P1283008268".
    """
    end = float(end) if end else time.time()
    events = _interval_split(interval)
    result = []

    while events:
        event, resumed = events.pop()

        if event != "R":
            # Ignore other events
            continue
        resumed = max(resumed, start or resumed)

        if events:  # Further events?
            if not events[-1][0] == "P":
                continue  # If not followed by "P", it's not a valid interval
            _, paused = events.pop()
            paused = min(paused, end)
        else:
            # Currently active, ends at time window
            paused = end

        if resumed >= paused:
            # Ignore empty intervals
            continue

        result.append(paused - resumed)

    return sum(result) if result else None


def _fmt_duration(duration):
    """Format duration value."""
    return fmt.human_duration(duration, 0, 2, True)


def _fmt_tags(tagset: Set[str]) -> str:
    """Convert set of strings to sorted space-separated list as a string."""
    return " ".join(sorted(tagset))


def _fmt_files(filelist) -> str:
    """Produce a file listing."""
    depth = max(i.path.count("/") for i in filelist)
    pad = ["\uFFFE"] * depth

    base_indent = " " * 38
    indent = 0
    result = []
    prev_path = pad
    sorted_files = sorted(
        (i.path.split("/")[:-1] + pad, i.path.rsplit("/", 1)[-1], i) for i in filelist
    )

    for path, name, fileinfo in sorted_files:
        path = path[:depth]
        if path != prev_path:
            common = min(
                [depth]
                + [
                    idx
                    for idx, (dirname, prev_name) in enumerate(zip(path, prev_path))
                    if dirname != prev_name
                ]
            )
            while indent > common:
                indent -= 1
                result.append(f"{base_indent}{' ' * indent}/")

            for dirname in path[common:]:
                if dirname == "\uFFFE":
                    break
                result.append(f"{base_indent}{' ' * indent}\\ {dirname}")
                indent += 1

        prio = {0: "off ", 1: "    ", 2: "high"}.get(fileinfo.prio, "????")
        result.append(
            f"  {prio} {fmt.iso_datetime(fileinfo.mtime)} {fmt.human_size(fileinfo.size)} {' ' * indent}| {name}"
        )

        prev_path = path

    while indent > 0:
        indent -= 1
        result.append(f"{base_indent}{' ' * indent}/")
    result.append(f"{base_indent}= {len(filelist)} file(s)")

    return "\n".join(result)


def detect_traits(item):
    """Build traits list from attributes of the passed item. Currently,
    "kind_51", "name" and "alias" are considered.

    See pyrosimple.util.traits:dectect_traits for more details.
    """
    return traits.detect_traits(
        name=item.name,
        alias=item.alias,
        filetype=(list(item.fetch("kind_51")) or [None]).pop(),
    )


class FieldDefinition:
    """Download item field."""

    CONSTANT_FIELDS = {"hash"}

    def __init__(
        self,
        valtype,
        name: str,
        doc: str,
        accessor=None,
        matcher=None,
        formatter=None,
        requires=None,
        prefilter_field=None,
    ):
        self.valtype = valtype
        self.name = name
        self.__doc__ = doc
        self.requires = requires or []
        self._accessor = accessor
        self._matcher = matcher
        self.formatter = formatter
        self.prefilter_field: Optional[str] = prefilter_field
        if accessor is None:
            self._accessor = lambda o: o.rpc_call("d." + name)
            if requires is None:
                self.requires = ["d." + name]

        if name in FIELD_REGISTRY:
            raise RuntimeError("INTERNAL ERROR: Duplicate field definition")
        FIELD_REGISTRY[name] = self

    def __repr__(self):
        """Return a representation of internal state."""
        return f"<{self.__class__.__name__}({self.valtype!r}, {self.name!r}, {self.__doc__!r})>"

    def __get__(self, obj, cls=None):
        if obj is None:
            return self
        return self.valtype(
            self._accessor(obj) if self._accessor else obj._fields[self.name]
        )

    def __delete__(self, obj):
        raise RuntimeError(f"Can't delete field {self.name!r}")


class ImmutableField(FieldDefinition):
    """Read-only download item field."""

    def __set__(self, obj, val):
        raise RuntimeError(f"Immutable field {self.name!r}")


class ConstantField(ImmutableField):
    """Read-only download item field with constant value."""

    # This can be cached


class DynamicField(ImmutableField):
    """Read-only download item field with dynamic value."""

    # This cannot be cached


class OnDemandField(DynamicField):
    """Only exists for backwards compatiblity."""


class MutableField(FieldDefinition):
    """Writable download item field"""

    def __init__(
        self,
        valtype,
        name,
        doc,
        accessor=None,
        matcher=None,
        formatter=None,
        requires=None,
        prefilter_field=None,
        setter: Callable = None,
    ):
        super().__init__(
            valtype, name, doc, accessor, matcher, formatter, requires, prefilter_field
        )
        self._setter = setter

    def __set__(self, obj, val, cls=None):
        if self._setter is None:
            raise NotImplementedError
        self._setter(obj, val)


FIELD_REGISTRY: Dict[str, FieldDefinition] = {}
FIELD_GENERATOR_REGISTRY: Dict[str, Callable] = {}


def field_lookup(name: str) -> Optional[FieldDefinition]:
    """Try to find field C{name}.

    @return: Field descriptions, see C{matching.ConditionParser} for details.
    """
    if name not in FIELD_REGISTRY:
        TorrentProxy.add_manifold_attribute(name)
    try:
        return FIELD_REGISTRY[name]
    except KeyError:
        return None


def core_fields():
    """Generate built-in field definitions"""
    yield ConstantField(
        bool,
        "is_private",
        "private flag set (no DHT/PEX)?",
        matcher=matching.BoolFilter,
        formatter=lambda val: "PRV" if val else "PUB",
        requires=["d.is_private"],
        prefilter_field="d.is_private=",
    )
    # Classification
    yield DynamicField(
        set,
        "tagged",
        "has certain tags? (not related to the 'tagged' view)",
        matcher=matching.TaggedAsFilter,
        accessor=lambda o: set(o.rpc_call("d.custom", ["tags"]).lower().split()),
        formatter=_fmt_tags,
        requires=["d.custom=tags"],
        prefilter_field="d.custom=tags",
    )
    yield DynamicField(
        set,
        "views",
        "views this item is attached to",
        matcher=matching.TaggedAsFilter,
        formatter=_fmt_tags,
        requires=["d.views"],
        prefilter_field="d.views=",
    )
    yield DynamicField(
        set,
        "kind",
        "ALL kinds of files in this item (the same as kind_0)",
        matcher=matching.TaggedAsFilter,
        formatter=_fmt_tags,
        accessor=lambda o: o.fetch("kind_0"),
    )
    yield DynamicField(
        list,
        "traits",
        "automatic classification of this item (audio, video, tv, movie, etc.)",
        matcher=matching.TaggedAsFilter,
        formatter=lambda v: "/".join(v or ["misc", "other"]),
        accessor=detect_traits,
    )

    # Basic fields
    yield ConstantField(
        str,
        "name",
        "name (file or root directory)",
        matcher=matching.PatternFilter,
        prefilter_field="d.name=",
    )
    yield ConstantField(
        int,
        "size",
        "data size",
        matcher=matching.ByteSizeFilter,
        accessor=lambda o: o.rpc_call("d.size_bytes"),
        requires=["d.size_bytes"],
        prefilter_field="d.size_bytes=",
    )
    yield MutableField(
        int,
        "prio",
        "priority (0=off, 1=low, 2=normal, 3=high)",
        matcher=matching.FloatFilter,
        accessor=lambda o: o.rpc_call("d.priority"),
        requires=["d.priority"],
        formatter=lambda val: "X- +"[val],
        prefilter_field="d.priority=",
    )
    yield ConstantField(
        str,
        "tracker",
        "first in the list of announce URLs",
        matcher=matching.PatternFilter,
        accessor=lambda o: (o.announce_urls(default=[None]) or [None])[0],
        requires=["t.multicall=,t.url=,t.is_enabled="],
    )

    def _alias_accessor(o):
        """Check the memoized alias custom field
        If it's defined as a key in the config, just return it
        Otherwise, check if it can be reduced down further.
        The worst case is that it doesn't exist to begin with,
        and then the tracker field needs to be referenced to figure it
        out properly.
        """
        memoized_alias = o.rpc_call("d.custom", ["memo_alias"])
        if memoized_alias in config.settings["ALIASES"]:
            return memoized_alias
        if memoized_alias:
            new_alias = config.map_announce2alias(memoized_alias)
            if new_alias:
                o.rpc_call("d.custom.set", ["memo_alias", new_alias])
            else:
                return memoized_alias
        else:
            new_alias = config.map_announce2alias(o.tracker)

        if memoized_alias != new_alias:
            o.rpc_call("d.custom.set", ["memo_alias", new_alias])
        return new_alias

    yield ConstantField(
        str,
        "alias",
        "tracker alias or domain",
        matcher=matching.PatternFilter,
        accessor=_alias_accessor,
        requires=["d.custom=memo_alias"],
    )
    yield DynamicField(
        str,
        "message",
        "current tracker message",
        matcher=matching.PatternFilter,
        requires=["d.message"],
        prefilter_field="d.message=",
    )

    # State
    yield DynamicField(
        bool,
        "is_open",
        "download open?",
        matcher=matching.BoolFilter,
        formatter=lambda val: "OPN" if val else "CLS",
        requires=["d.is_open"],
        prefilter_field="d.is_open=",
    )
    yield DynamicField(
        bool,
        "is_active",
        "download active?",
        matcher=matching.BoolFilter,
        formatter=lambda val: "ACT" if val else "STP",
        requires=["d.is_active"],
        prefilter_field="d.is_active=",
    )
    yield DynamicField(
        bool,
        "is_complete",
        "download complete?",
        matcher=matching.BoolFilter,
        formatter=lambda val: "DONE" if val else "PART",
        accessor=lambda o: o.rpc_call("d.complete"),
        requires=["d.complete"],
        prefilter_field="d.complete=",
    )
    yield ConstantField(
        bool,
        "is_multi_file",
        "single- or multi-file download?",
        matcher=matching.BoolFilter,
        formatter=lambda val: "DIR " if val else "FILE",
        prefilter_field="d.is_multi_file=",
    )
    yield MutableField(
        bool,
        "is_ignored",
        "ignore commands?",
        matcher=matching.BoolFilter,
        formatter=lambda val: "IGN!" if int(val) else "HEED",
        accessor=lambda o: o.rpc_call("d.ignore_commands"),
        requires=["d.ignore_commands"],
        prefilter_field="d.ignore_commands=",
        setter=lambda o, val: o.ignore(int(val)),
    )
    yield DynamicField(
        bool,
        "is_ghost",
        "has no data file or directory?",
        matcher=matching.BoolFilter,
        accessor=lambda o: not os.path.exists(o.datapath()) if o.datapath() else None,
        formatter=lambda val: "GHST" if val else "DATA",
        requires=["d.directory", "d.is_multi_file"],
    )

    # Paths
    yield DynamicField(
        str,
        "directory",
        "directory containing download data",
        matcher=matching.PatternFilter,
        prefilter_field="d.directory=",
    )
    yield DynamicField(
        str,
        "path",
        "path to download data",
        matcher=matching.PatternFilter,
        accessor=lambda o: o.datapath(),
        requires=["d.directory", "d.is_multi_file"],
    )
    yield DynamicField(
        str,
        "realpath",
        "real path to download data",
        matcher=matching.PatternFilter,
        accessor=lambda o: os.path.realpath(o.datapath()),
        requires=["d.directory", "d.is_multi_file"],
    )
    yield ConstantField(
        str,
        "metafile",
        "path to torrent file",
        matcher=matching.PatternFilter,
        accessor=lambda o: os.path.expanduser(str(o.rpc_call("d.tied_to_file"))),
        requires=["d.tied_to_fiel"],
    )
    yield ConstantField(
        str,
        "sessionfile",
        "path to session file",
        matcher=matching.PatternFilter,
        accessor=lambda o: os.path.expanduser(str(o.rpc_call("d.session_file"))),
        requires=["d.session_file"],
        prefilter_field="d.session_file=",
    )
    yield ConstantField(
        list,
        "files",
        "list of files in this item",
        matcher=matching.FilesFilter,
        formatter=_fmt_files,
        accessor=lambda o: o._get_files(),
    )
    yield ConstantField(
        int,
        "fno",
        "number of files in this item",
        matcher=matching.FloatFilter,
        accessor=lambda o: o.rpc_call("d.size_files"),
        requires=["d.size_files"],
    )

    # Bandwidth & Data Transfer
    yield DynamicField(
        percent,
        "done",
        "completion in percent",
        matcher=matching.FloatFilter,
        accessor=lambda o: float(o.rpc_call("d.completed_bytes"))
        / o.rpc_call("d.size_bytes"),
        requires=["d.size_bytes", "d.completed_bytes"],
    )
    yield DynamicField(
        ratio_float,
        "ratio",
        "normalized ratio (1:1 = 1.0)",
        matcher=matching.FloatFilter,
        accessor=lambda o: o.rpc_call("d.ratio"),
        requires=["d.ratio"],
        prefilter_field="d.ratio=",
    )
    yield DynamicField(
        int,
        "uploaded",
        "amount of uploaded data",
        matcher=matching.ByteSizeFilter,
        accessor=lambda o: o.rpc_call("d.up.total"),
        requires=["d.up.total"],
        prefilter_field="d.up.total=",
    )
    yield DynamicField(
        int,
        "xfer",
        "transfer rate",
        matcher=matching.ByteSizeFilter,
        accessor=lambda o: o.fetch("up") + o.fetch("down"),
        requires=["d.up.rate", "d.down.rate"],
    )

    def _last_xfer_accessor(o):
        engine = o._engine
        if config.settings.SAFETY_CHECKS_ENABLED and not engine.has_method(
            "d.timestamp.last_xfer"
        ):
            warnings.warn(
                "Method 'd.timestamp.last_xfer' does not exist! See https://kannibalox.github.io/pyrosimple/rtorrent-config/ for information on setting up rtorrent.rc.",
                stacklevel=2,
            )
            return 0
        else:
            return int(o.rpc_call("d.timestamp.last_xfer") or 0)

    yield DynamicField(
        int,
        "last_xfer",
        "last time data was transferred",
        matcher=matching.TimeFilter,
        accessor=_last_xfer_accessor,
        formatter=fmt.iso_datetime_optional,
    )
    yield DynamicField(
        int,
        "down",
        "download rate",
        matcher=matching.ByteSizeFilter,
        accessor=lambda o: o.rpc_call("d.down.rate"),
        requires=["d.down.rate"],
        prefilter_field="d.down.rate=",
    )
    yield DynamicField(
        int,
        "up",
        "upload rate",
        matcher=matching.ByteSizeFilter,
        accessor=lambda o: o.rpc_call("d.up.rate"),
        requires=["d.up.rate"],
        prefilter_field="d.up.rate=",
    )
    yield DynamicField(
        str,
        "throttle",
        "throttle group name (NULL=unlimited, NONE=global)",
        matcher=matching.PatternFilter,
        accessor=lambda o: o.rpc_call("d.throttle_name"),
        formatter=lambda v: v if v else "NONE",
        requires=["d.throttle_name"],
        prefilter_field="d.throttle_name=",
    )

    # Lifecyle
    yield DynamicField(
        int,
        "loaded",
        "time metafile was loaded",
        matcher=matching.TimeFilterNotNull,
        accessor=lambda o: int(o.rpc_call("d.custom", ["tm_loaded"]) or "0", 10),
        formatter=fmt.iso_datetime_optional,
        requires=["d.custom=tm_loaded"],
        prefilter_field="d.custom=tm_loaded",
    )
    yield DynamicField(
        int,
        "started",
        "time download was FIRST started",
        matcher=matching.TimeFilterNotNull,
        accessor=lambda o: int(o.rpc_call("d.custom", ["tm_started"]) or "0", 10),
        formatter=fmt.iso_datetime_optional,
        requires=["d.custom=tm_started"],
        prefilter_field="d.custom=tm_loaded",
    )

    def _leechtime_accessor(o):
        leechtime = _interval_sum(
            o.rpc_call("d.custom", ["activations"]), end=o.completed
        )
        if not leechtime:
            leechtime = _duration(o.started, o.completed)
        return leechtime

    yield DynamicField(
        untyped,
        "leechtime",
        "time taken from start to completion",
        matcher=matching.DurationFilter,
        accessor=_leechtime_accessor,
        formatter=_fmt_duration,
        requires=[
            "d.custom=tm_completed",
            "d.custom=tm_started",
            "d.custom=activations",
        ],
    )
    yield DynamicField(
        int,
        "completed",
        "time download was finished",
        matcher=matching.TimeFilterNotNull,
        accessor=lambda o: int(o.rpc_call("d.custom", ["tm_completed"]) or "0", 10),
        formatter=fmt.iso_datetime_optional,
        requires=["d.custom=tm_completed"],
        prefilter_field="d.custom=tm_completed",
    )
    yield DynamicField(
        untyped,
        "seedtime",
        "total seeding time after completion",
        matcher=matching.DurationFilter,
        accessor=lambda o: _interval_sum(
            o.rpc_call("d.custom", ["activations"]), start=o.completed
        )
        if o.rpc_call("d.complete")
        else None,
        formatter=_fmt_duration,
        requires=["d.custom=tm_completed", "d.complete", "d.custom=activations"],
    )

    def _last_active_accessor(o):
        engine = o._engine
        if config.settings.SAFETY_CHECKS_ENABLED and not engine.has_method(
            "d.timestamp.last_active"
        ):
            warnings.warn(
                "Method 'd.timestamp.last_active' does not exist! See https://kannibalox.github.io/pyrosimple/rtorrent-config/ for information on setting up rtorrent.rc.",
                stacklevel=2,
            )
            return 0
        else:
            return int(o.rpc_call("d.timestamp.last_active") or 0)

    yield DynamicField(
        int,
        "active",
        "last time a peer was connected",
        matcher=matching.TimeFilter,
        accessor=_last_active_accessor,
        formatter=fmt.iso_datetime_optional,
        requires=["d.timestamp.last_active"],
    )
    yield DynamicField(
        int,
        "stopped",
        "time download was last stopped or paused",
        matcher=matching.TimeFilterNotNull,
        accessor=lambda o: (
            _interval_split(o.rpc_call("d.custom", ["activations"]), only="P")
            + [(0, 0)]
        )[0][1],
        formatter=fmt.iso_datetime_optional,
        requires=["d.custom=activations"],
    )


def generate_custom_field(name: str) -> FieldDefinition:
    """Create fields from custom keys"""
    custom_name = name.split("_", 1)[1]
    accessor = lambda o: o.rpc_call("d.custom", [custom_name])
    description = f"custom attribute {custom_name}"
    requires = [f"d.custom={custom_name}"]
    # Handle custom1, custom2, etc as a special case
    if len(custom_name) == 1 and custom_name in "12345":
        accessor = lambda o: o.rpc_call(f"d.custom{custom_name}")
        description = f"custom{custom_name}"
        requires = [f"d.custom{custom_name}"]
    return DynamicField(
        str,
        name,
        description,
        matcher=matching.PatternFilter,
        accessor=accessor,
        requires=requires,
    )


def generate_kind_field(name: str) -> FieldDefinition:
    """Generate kind percentile fields"""
    limit = int(name[5:].lstrip("0") or "0", 10)
    if limit > 100:
        raise error.UserError(f"kind_N: N can't be greater than 100 in {name!r}")
    return DynamicField(
        set,
        name,
        f"kinds of files that make up more than {limit}% of this item's size",
        accessor=lambda o: o._get_kind(limit),
        matcher=matching.TaggedAsFilter,
        formatter=_fmt_tags,
        requires=["d.custom=kind"],
    )


def generate_guessit_field(name: str) -> Optional[FieldDefinition]:
    """Create fields based on the guessit parser (if installed)."""
    try:
        import guessit  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError:
        logger.error(
            "'guessit' module not found while loading field '%s', install with: pip install guessit",
            name,
        )
        return None

    guess_field = name[8:]

    def _guess_accessor(o):
        return guessit.guessit(o.rpc_call("d.name")).get(guess_field, "")

    return DynamicField(
        str,
        name,
        f"Guessit field {guess_field}",
        accessor=_guess_accessor,
        requires=["d.name"],
    )


class TorrentProxy:
    """A single download item."""

    @classmethod
    def add_manifold_attribute(cls, name) -> Optional[FieldDefinition]:
        """Register a manifold engine attribute.

        @return: field definition object, or None if "name" isn't a manifold attribute.
        """
        if name in FIELD_REGISTRY:
            return FIELD_REGISTRY[name]
        for prefix, generator in FIELD_GENERATOR_REGISTRY.items():
            if name.startswith(prefix):
                field = generator(name)
                if field is not None:
                    setattr(cls, name, field)
                    FIELD_REGISTRY[name] = field
                    return cast(FieldDefinition, field)
                else:
                    return None
        return None

    @classmethod
    def add_field_generator(cls, prefix: str, generator: Callable):
        """Add a field generator with a given prefix to the registry"""
        FIELD_GENERATOR_REGISTRY[prefix] = generator

    @classmethod
    def add_field(cls, field):
        """Add a custom field to the class"""
        setattr(cls, field.name, field)
        FIELD_REGISTRY[field.name] = field
        if isinstance(field, ConstantField):
            FieldDefinition.CONSTANT_FIELDS.add(field.name)

    @classmethod
    def add_core_fields(cls, *_, **__):
        """Add any custom fields defined in the configuration."""
        for field in core_fields():
            cls.add_field(field)
        for prefix, generator in {
            "custom_": generate_custom_field,
            "kind_": generate_kind_field,
            "guessit_": generate_guessit_field,
        }.items():
            cls.add_field_generator(prefix, generator)

    def __init__(self):
        """Initialize object."""
        self._fields = {}

    def __hash__(self):
        """Make item hashable for Python."""
        return self.hash

    def __eq__(self, other):
        """Compare items based on their infohash."""
        return other and self.hash == getattr(other, "hash", None)

    hash = ConstantField(
        str,
        "hash",
        "info hash",
        matcher=matching.PatternFilter,
        prefilter_field="d.hash=",
    )

    def __repr__(self):
        """Return a representation of internal state."""

        def mask(key, val):
            "helper to hide sensitive stuff"
            if key in ("tracker", "custom_memo_alias"):
                return key, metafile.mask_keys(val)
            return key, val

        return "<{}({})>".format(
            self.__class__.__name__,
            ", ".join(sorted("%s=%r" % mask(k, v) for k, v in self._fields.items())),
        )


TorrentProxy.add_core_fields()


class TorrentView:
    """A view on a subset of torrent items."""

    def __init__(self, engine, viewname, matcher=None):
        """Initialize view on torrent items."""
        self.engine = engine
        self.viewname = viewname or "default"
        self.matcher = matcher
        self._items = None

    def __iter__(self):
        return self.items()

    def _fetch_items(self):
        """Fetch to attribute."""
        if self._items is None:
            self._items = list(self.engine.items(self))

        return self._items

    def _check_hash_view(self) -> Optional[str]:
        """Return infohash if view name refers to a single item, else None."""
        infohash = None
        if self.viewname.startswith("#"):
            infohash = self.viewname[1:]
        elif len(self.viewname) == 40:
            try:
                int(self.viewname, 16)
            except (TypeError, ValueError):
                pass
            else:
                infohash = str(self.viewname)
        return infohash

    def size(self) -> int:
        """Total unfiltered size of view."""
        if self._check_hash_view():
            return 1
        return int(self.engine.open().view.size(rpc.NOHASH, self.viewname))

    def items(self):
        """Get list of download items."""
        if self.matcher:
            for item in self._fetch_items():
                if self.matcher.match(item):
                    yield item
        else:
            for item in self._fetch_items():
                yield item



""" rTorrent Watch Jobs.

    Copyright (c) 2012 The PyroScope Project <pyroscope.project@gmail.com>
"""


import asyncio
import logging
import os
import time

from pathlib import Path
from typing import Any, Dict

import pyrosimple

from pyrosimple import config as configuration
from pyrosimple import error
from pyrosimple.scripts.base import ScriptBase, ScriptBaseWithConfig
from pyrosimple.torrent import rtorrent
from pyrosimple.util import metafile, pymagic, rpc
from pyrosimple.util.parts import Bunch


try:
    import pyinotify
except ImportError:
    pyinotify = Bunch(WatchManager=None, ProcessEvent=object)


class MetafileHandler:
    """Handler for loading metafiles into rTorrent."""

    def __init__(self, job, pathname: str):
        """Create a metafile handler."""
        self.metadata: metafile.Metafile
        self.job = job
        self.pathname = Path(pathname).resolve()
        self.ns: Dict[str, Any] = {
            "pathname": os.path.abspath(pathname),
            "info_hash": None,
            "tracker_alias": None,
        }

    def parse(self) -> bool:
        """Parse metafile and check pre-conditions."""
        try:
            if not os.path.getsize(self.pathname):
                # Ignore 0-byte dummy files (Firefox creates these while downloading)
                self.job.LOG.warning("Ignoring 0-byte metafile '%s'", self.pathname)
                return False

            self.metadata = metafile.Metafile.from_file(self.pathname)
            self.metadata.check_meta()
        except OSError as exc:
            self.job.LOG.error(
                "Can't read metafile '%s' (%s)",
                self.pathname,
                str(exc).replace(f": '{self.pathname}'", ""),
            )
            return False
        except ValueError as exc:
            self.job.LOG.error("Invalid metafile '%s': %s", self.pathname, exc)
            return False

        self.ns["info_hash"] = self.metadata.info_hash()
        self.ns["info_name"] = self.metadata["info"]["name"]
        self.job.LOG.info(
            "Loaded '%s' from metafile '%s'", self.metadata.info_hash(), self.pathname
        )

        # Check whether item is already loaded
        try:
            name = self.job.proxy.d.name(self.metadata.info_hash())
        except rpc.HashNotFound:
            pass
        except rpc.RpcError as exc:
            if exc.faultString != "Could not find info-hash.":
                self.job.LOG.error(
                    "While checking for #%s: %s", self.metadata.info_hash(), exc
                )
                return False
        else:
            self.job.LOG.warn(
                "Item #%s '%s' already added to client", self.metadata.info_hash(), name
            )
            if (
                self.job.config.get("remove_already_added", False)
                and not self.job.config["dry_run"]
            ):
                Path(self.pathname).unlink()
            return False

        return True

    def addinfo(self) -> None:
        """Add known facts to templating namespace."""
        # Basic values
        self.ns["watch_path"] = self.job.config["path"]
        self.ns["relpath"] = None
        for watch in self.job.config["path"]:
            path = Path(self.pathname)
            try:
                self.ns["relpath"] = path.relative_to(watch)
                break
            except ValueError:
                pass

        # Build indicator flags for target state from filename
        flags = str(self.pathname).split(os.sep)
        flags.extend(flags[-1].split("."))
        self.ns["flags"] = {i for i in flags if i}

        # Metafile stuff
        announce = self.metadata.get("announce", None)
        if announce:
            self.ns["tracker_alias"] = configuration.map_announce2alias(announce)

        main_file = self.ns["info_name"]
        if "files" in self.metadata["info"]:
            main_file = list(
                sorted(
                    (i["length"], i["path"][-1]) for i in self.metadata["info"]["files"]
                )
            )[-1][1]
        self.ns["filetype"] = os.path.splitext(main_file)[1]

        # Finally, expand commands from templates
        self.ns["commands"] = []
        for key, cmd in sorted(self.job.custom_cmds.items()):
            try:
                template = rtorrent.env.from_string(cmd)
                for split_cmd in rtorrent.format_item(
                    template, {}, defaults=self.ns
                ).split():
                    self.ns["commands"].append(split_cmd.strip())
            except error.LoggableError as exc:
                self.job.LOG.error(f"While expanding '{key}' custom command: {exc}")

    def load(self) -> None:
        """Load metafile into client."""
        if not self.metadata.info_hash() and not self.parse():
            return

        self.addinfo()

        try:
            # TODO: Scrub metafile if requested

            # Determine target state
            start_it = self.job.config.get("load_mode", "").lower() in (
                "start",
                "started",
            )

            if "start" in self.ns["flags"]:
                start_it = True
            elif "load" in self.ns["flags"]:
                start_it = False

            # Load metafile into client
            load_cmd = self.job.proxy.load.verbose
            if start_it:
                load_cmd = self.job.proxy.load.start_verbose

            self.job.LOG.debug(
                "Templating values are:\n    %s"
                % "\n    ".join(
                    f"{key}={repr(val)}" for key, val in sorted(self.ns.items())
                )
            )

            if self.job.config["dry_run"]:
                self.job.LOG.info(
                    f"Would load: {self.pathname} with commands {self.ns['commands']}"
                )
                return

            self.job.LOG.debug(
                f"Loading {self.pathname} with commands {self.ns['commands']}"
            )

            load_cmd(rpc.NOHASH, str(self.pathname), *tuple(self.ns["commands"]))
            time.sleep(0.05)  # let things settle

            # Announce new item
            if self.job.config["print_to_client"]:
                try:
                    name = self.job.proxy.d.name(self.metadata.info_hash())
                except rpc.HashNotFound:
                    name = "NOHASH"
                msg = "{}: Loaded '{}' from '{}/' {}".format(
                    self.job.__class__.__name__,
                    name,
                    os.path.dirname(self.pathname).rstrip(os.sep),
                    "[started]" if start_it else "[normal]",
                )
                self.job.proxy.log(rpc.NOHASH, msg)

            # TODO: Evaluate fields and set client values
            # TODO: Add metadata to tied file if requested

            # TODO: Execute commands AFTER adding the item, with full templating
            # Example: Labeling - add items to a persistent view, i.e. "postcmd = view.set_visible={{label}}"
            #   could also be done automatically from the path, see above under "flags" (autolabel = True)
            #   and add traits to the flags, too, in that case

        except rpc.ERRORS as exc:
            self.job.LOG.error("While loading #%s: %s", self.metadata.info_hash(), exc)

    def handle(self):
        """Handle metafile."""
        if self.parse():
            self.load()


class RemoteWatch:
    """rTorrent remote torrent file watch."""

    def __init__(self, config=None):
        """Set up remote watcher."""
        self.config = config or {}
        self.LOG = pymagic.get_class_logger(self)
        self.LOG.debug("Remote watcher created with config %r", self.config)

    def run(self):
        """Check remote watch target."""
        # TODO: ftp. ssh, and remote rTorrent instance (extra view?) as sources!
        # config:
        #   local_dir   storage path (default local sessiondir + '/remote-watch-' + jobname
        #   target      URL of target to watch


class TreeWatchHandler(pyinotify.ProcessEvent):
    """inotify event handler for rTorrent folder tree watch.

    See https://github.com/seb-m/pyinotify/.
    """

    METAFILE_EXT = (".torrent", ".load", ".start", ".queue")

    def handle_path(self, event):
        """Handle a path-related event."""
        self.job.LOG.debug(f"Notification {event!r}")
        if event.dir:
            return

        if any(event.pathname.endswith(i) for i in self.METAFILE_EXT):
            MetafileHandler(self.job, event.pathname).handle()
        elif os.path.basename(event.pathname) == "watch.ini":
            self.job.LOG.info(f"NOT YET Reloading watch config for '{event.path}'")
            # TODO: Load new metadata

    def process_IN_CLOSE_WRITE(self, event):
        """File written."""
        # <Event dir=False name=xx path=/var/torrent/watch/tmp pathname=/var/torrent/watch/tmp/xx>
        self.handle_path(event)

    def process_IN_MOVED_TO(self, event):
        """File moved into tree."""
        # <Event dir=False name=yy path=/var/torrent/watch/tmp pathname=/var/torrent/watch/tmp/yy>
        self.handle_path(event)

    def process_default(self, event):
        """Fallback."""
        if self.job.LOG.isEnabledFor(logging.DEBUG):
            # On debug level, we subscribe to ALL events, so they're expected in that case ;)
            if self.job.config["trace_inotify"]:
                self.job.LOG.debug(f"Ignored inotify event:\n    {event!r}")
        else:
            self.job.LOG.warning(f"Unexpected inotify event {event!r}")


class TreeWatch:
    """rTorrent folder tree watch via inotify."""

    def __init__(self, config=None):
        self.config = config or {}
        self.LOG = pymagic.get_class_logger(self)
        if "log_level" in self.config:
            self.LOG.setLevel(config["log_level"])
        self.LOG.debug("Tree watcher created with config %r", self.config)
        self.config.setdefault("print_to_client", True)
        self.config.setdefault("dry_run", False)
        self.config.setdefault("started", False)
        self.config.setdefault("trace_inotify", False)
        self.config.setdefault("check_unhandled", False)
        self.config.setdefault("remove_unhandled", False)
        self.config.setdefault("remove_already_added", False)

        self.manager = None
        self.handler = None
        self.notifier = None

        if "path" not in self.config:
            raise error.UserError("You need to set 'path' in the configuration!")

        self.config["path"] = {
            Path(p).expanduser().absolute()
            for p in self.config["path"].split(os.pathsep)
        }
        for path in self.config["path"]:
            if not path.is_dir():
                raise error.UserError(f"Path '{path}' is not a directory!")

        # Assemble custom commands
        self.custom_cmds = {}
        for key, val in self.config.items():
            if key.startswith("cmd_"):
                self.custom_cmds[key] = val

        # Get client proxy
        url = None
        if "scgi_url" in self.config:
            url = pyrosimple.config.lookup_connection_alias(
                str(self.config.get("scgi_url"))
            )
        self.engine = pyrosimple.connect(url)
        self.proxy = self.engine.open()

        self.setup()

    def setup(self):
        """Set up inotify manager.

        See https://github.com/seb-m/pyinotify/.
        """
        if not pyinotify.WatchManager:
            raise error.UserError(
                f"You need to install 'pyinotify' to use {self.__class__.__name__}!"
            )

        self.manager = pyinotify.WatchManager()
        self.handler = TreeWatchHandler(job=self)
        self.notifier = pyinotify.AsyncNotifier(self.manager, self.handler)

        if self.LOG.isEnabledFor(logging.DEBUG):
            mask = pyinotify.ALL_EVENTS
        else:
            mask = (
                pyinotify.IN_CLOSE_WRITE  # pylint: disable=no-member
                | pyinotify.IN_MOVED_TO  # pylint: disable=no-member
            )

        # Add all configured base dirs
        for path in self.config["path"]:
            self.manager.add_watch(path, mask, rec=True, auto_add=True)

    def run(self):
        """Regular maintenance and fallback task."""
        if self.config.get("check_unhandled", False):
            for path in self.config["path"]:
                for filepath in Path(path).rglob("**/*.torrent"):
                    MetafileHandler(self, filepath).handle()
                    if (
                        self.config.get("remove_unhandled", False)
                        and filepath.exists()
                        and not self.config["dry_run"]
                    ):
                        filepath.unlink()


class TreeWatchCommand(ScriptBaseWithConfig):
    """
    Use tree watcher directly from cmd line, call it like this:
        python -m pyrosimple.torrent.watch <DIR>

    If the argument is a file, the templating namespace for that metafile is
    dumped (for testing and debugging purposes).
    """

    # log level for user-visible standard logging
    STD_LOG_LEVEL = logging.DEBUG

    # argument description for the usage information
    ARGS_HELP = "<directory>"

    def mainloop(self):
        """The main loop."""
        # Print usage if not enough args or bad options
        if len(self.args) < 1:
            self.parser.error(
                "You have to provide the root directory of your watch tree, or a metafile path!"
            )

        pathname = os.path.abspath(self.args[0])
        if os.path.isdir(pathname):
            watch = TreeWatch(
                Bunch(
                    path=pathname,
                    job_name="watch",
                    active=True,
                    dry_run=True,
                    load_mode=None,
                )
            )
            asyncio.sleep(0)
        else:
            config = Bunch()
            config.update(
                {
                    key.split(".", 2)[-1]: val
                    for key, val in configuration.settings.TORQUE.items()
                    if key.startswith("job.treewatch.")
                }
            )
            config.update(
                dict(
                    path=os.path.dirname(os.path.dirname(pathname)),
                    job_name="treewatch",
                    active=False,
                    dry_run=True,
                )
            )
            watch = TreeWatch(config)
            handler = MetafileHandler(watch, pathname)

            ok = handler.parse()
            self.LOG.debug(
                "Metafile '%s' would've %sbeen loaded", pathname, ("" if ok else "NOT ")
            )

            if ok:
                handler.addinfo()
                self.LOG.info(
                    "Templating values are:\n    %s",
                    "\n    ".join(
                        f"{key}={repr(val)}" for key, val in sorted(handler.ns.items())
                    ),
                )

    @classmethod
    def main(cls):
        """The entry point."""
        ScriptBase.setup()
        cls().run()


if __name__ == "__main__":
    TreeWatchCommand.main()
